@@ -1,5 +1,5 @@
 /* Extended Module Player
- * Copyright (C) 1996-2021 Claudio Matsuoka and Hipolito Carraro Jr
+ * Copyright (C) 1996-2022 Claudio Matsuoka and Hipolito Carraro Jr
  *
  * Permission is hereby granted, free of charge, to any person obtaining a
  * copy of this software and associated documentation files (the "Software"),
@@ -20,16 +20,15 @@
  * THE SOFTWARE.
  */
 
-#include <sys/stat.h>
 #include <errno.h>
 
 #include "format.h"
 #include "list.h"
 #include "hio.h"
-#include "tempfile.h"
 #include "loaders/loader.h"
 
 #ifndef LIBXMP_NO_DEPACKERS
+#include "tempfile.h"
 #include "depackers/depacker.h"
 #endif
 
@@ -39,17 +38,13 @@
 #endif
 
 
-extern struct format_loader *format_loaders[];
-
 void libxmp_load_prologue(struct context_data *);
 void libxmp_load_epilogue(struct context_data *);
 int  libxmp_prepare_scan(struct context_data *);
 
 #ifndef LIBXMP_CORE_PLAYER
 #define BUFLEN 16384
-#endif
 
-#ifndef LIBXMP_CORE_PLAYER
 static void set_md5sum(HIO_HANDLE *f, unsigned char *digest)
 {
 	unsigned char buf[BUFLEN];
@@ -68,18 +63,18 @@ static void set_md5sum(HIO_HANDLE *f, unsigned char *digest)
 static char *get_dirname(const char *name)
 {
 	char *dirname;
-	const char *div;
+	const char *p;
 	ptrdiff_t len;
 
-	if ((div = strrchr(name, '/')) != NULL) {
-		len = div - name + 1;
-		dirname = malloc(len + 1);
+	if ((p = strrchr(name, '/')) != NULL) {
+		len = p - name + 1;
+		dirname = (char *) malloc(len + 1);
 		if (dirname != NULL) {
 			memcpy(dirname, name, len);
 			dirname[len] = 0;
 		}
 	} else {
-		dirname = strdup("");
+		dirname = libxmp_strdup("");
 	}
 
 	return dirname;
@@ -87,13 +82,13 @@ static char *get_dirname(const char *name)
 
 static char *get_basename(const char *name)
 {
-	const char *div;
+	const char *p;
 	char *basename;
 
-	if ((div = strrchr(name, '/')) != NULL) {
-		basename = strdup(div + 1);
+	if ((p = strrchr(name, '/')) != NULL) {
+		basename = libxmp_strdup(p + 1);
 	} else {
-		basename = strdup(name);
+		basename = libxmp_strdup(name);
 	}
 
 	return basename;
@@ -115,7 +110,7 @@ static int test_module(struct xmp_test_info *info, HIO_HANDLE *h)
 		if (format_loaders[i]->test(h, buf, 0) == 0) {
 			int is_prowizard = 0;
 
-#if !defined(LIBXMP_CORE_PLAYER) && !defined(LIBXMP_NO_PROWIZARD)
+#ifndef LIBXMP_NO_PROWIZARD
 			if (strcmp(format_loaders[i]->name, "prowizard") == 0) {
 				hio_seek(h, 0, SEEK_SET);
 				pw_test_format(h, buf, 0, info);
@@ -140,16 +135,17 @@ static int test_module(struct xmp_test_info *info, HIO_HANDLE *h)
 int xmp_test_module(const char *path, struct xmp_test_info *info)
 {
 	HIO_HANDLE *h;
-	struct stat st;
-	int ret;
 #ifndef LIBXMP_NO_DEPACKERS
 	char *temp = NULL;
 #endif
+	int ret;
 
-	if (stat(path, &st) < 0)
+	ret = libxmp_get_filetype(path);
+
+	if (ret == XMP_FILETYPE_NONE) {
 		return -XMP_ERROR_SYSTEM;
-
-	if (S_ISDIR(st.st_mode)) {
+	}
+	if (ret & XMP_FILETYPE_DIR) {
 		errno = EISDIR;
 		return -XMP_ERROR_SYSTEM;
 	}
@@ -158,16 +154,10 @@ int xmp_test_module(const char *path, struct xmp_test_info *info)
 		return -XMP_ERROR_SYSTEM;
 
 #ifndef LIBXMP_NO_DEPACKERS
-	if (libxmp_decrunch(&h, path, &temp) < 0) {
+	if (libxmp_decrunch(h, path, &temp) < 0) {
 		ret = -XMP_ERROR_DEPACK;
 		goto err;
 	}
-
-	/* get size after decrunch */
-	if (hio_size(h) < 256) {	/* set minimum valid module size */
-		ret = -XMP_ERROR_FORMAT;
-		goto err;
-	}
 #endif
 
 	ret = test_module(info, h);
@@ -191,7 +181,7 @@ int xmp_test_module_from_memory(const void *mem, long size, struct xmp_test_info
 		return -XMP_ERROR_INVALID;
 	}
 
-	if ((h = hio_open_mem(mem, size)) == NULL)
+	if ((h = hio_open_mem(mem, size, 0)) == NULL)
 		return -XMP_ERROR_SYSTEM;
 
 	ret = test_module(info, h);
@@ -212,16 +202,10 @@ int xmp_test_module_from_file(void *file, struct xmp_test_info *info)
 		return -XMP_ERROR_SYSTEM;
 
 #ifndef LIBXMP_NO_DEPACKERS
-	if (libxmp_decrunch(&h, NULL, &temp) < 0) {
+	if (libxmp_decrunch(h, NULL, &temp) < 0) {
 		ret = -XMP_ERROR_DEPACK;
 		goto err;
 	}
-
-	/* get size after decrunch */
-	if (hio_size(h) < 256) {	/* set minimum valid module size */
-		ret = -XMP_ERROR_FORMAT;
-		goto err;
-	}
 #endif
 
 	ret = test_module(info, h);
@@ -265,7 +249,6 @@ static int load_module(xmp_context opaque, HIO_HANDLE *h)
 	test_result = load_result = -1;
 	for (i = 0; format_loaders[i] != NULL; i++) {
 		hio_seek(h, 0, SEEK_SET);
-		hio_error(h); /* reset error flag */
 
 		D_(D_WARN "test %s", format_loaders[i]->name);
 		test_result = format_loaders[i]->test(h, NULL, 0);
@@ -277,11 +260,6 @@ static int load_module(xmp_context opaque, HIO_HANDLE *h)
 		}
 	}
 
-#ifndef LIBXMP_CORE_PLAYER
-	if (test_result == 0 && load_result == 0)
-		set_md5sum(h, m->md5);
-#endif
-
 	if (test_result < 0) {
 		xmp_release_module(opaque);
 		return -XMP_ERROR_FORMAT;
@@ -330,6 +308,11 @@ static int load_module(xmp_context opaque, HIO_HANDLE *h)
 		libxmp_adjust_string(mod->xxs[i].name);
 	}
 
+#ifndef LIBXMP_CORE_PLAYER
+	if (test_result == 0 && load_result == 0)
+		set_md5sum(h, m->md5);
+#endif
+
 	libxmp_load_epilogue(ctx);
 
 	ret = libxmp_prepare_scan(ctx);
@@ -358,22 +341,21 @@ int xmp_load_module(xmp_context opaque, const char *path)
 	struct context_data *ctx = (struct context_data *)opaque;
 #ifndef LIBXMP_CORE_PLAYER
 	struct module_data *m = &ctx->m;
-	long size;
 #endif
 #ifndef LIBXMP_NO_DEPACKERS
 	char *temp_name;
 #endif
 	HIO_HANDLE *h;
-	struct stat st;
 	int ret;
 
 	D_(D_WARN "path = %s", path);
 
-	if (stat(path, &st) < 0) {
+	ret = libxmp_get_filetype(path);
+
+	if (ret == XMP_FILETYPE_NONE) {
 		return -XMP_ERROR_SYSTEM;
 	}
-
-	if (S_ISDIR(st.st_mode)) {
+	if (ret & XMP_FILETYPE_DIR) {
 		errno = EISDIR;
 		return -XMP_ERROR_SYSTEM;
 	}
@@ -384,20 +366,12 @@ int xmp_load_module(xmp_context opaque, const char *path)
 
 #ifndef LIBXMP_NO_DEPACKERS
 	D_(D_INFO "decrunch");
-	if (libxmp_decrunch(&h, path, &temp_name) < 0) {
+	if (libxmp_decrunch(h, path, &temp_name) < 0) {
 		ret = -XMP_ERROR_DEPACK;
 		goto err;
 	}
 #endif
 
-#ifndef LIBXMP_CORE_PLAYER
-	size = hio_size(h);
-	if (size < 256) {		/* get size after decrunch */
-		ret = -XMP_ERROR_FORMAT;
-		goto err;
-	}
-#endif
-
 	if (ctx->state > XMP_STATE_UNLOADED)
 		xmp_release_module(opaque);
 
@@ -415,7 +389,7 @@ int xmp_load_module(xmp_context opaque, const char *path)
 	}
 
 	m->filename = path;	/* For ALM, SSMT, etc */
-	m->size = size;
+	m->size = hio_size(h);
 #else
 	ctx->m.filename = NULL;
 	ctx->m.dirname = NULL;
@@ -452,7 +426,7 @@ int xmp_load_module_from_memory(xmp_context opaque, const void *mem, long size)
 		return -XMP_ERROR_INVALID;
 	}
 
-	if ((h = hio_open_mem(mem, size)) == NULL)
+	if ((h = hio_open_mem(mem, size, 0)) == NULL)
 		return -XMP_ERROR_SYSTEM;
 
 	if (ctx->state > XMP_STATE_UNLOADED)
@@ -579,17 +553,9 @@ void xmp_release_module(xmp_context opaque)
 	}
 
 	free(m->xtra);
+	free(m->midi);
 	m->xtra = NULL;
-
-#ifndef LIBXMP_CORE_DISABLE_IT
-	if (m->xsmp != NULL) {
-		for (i = 0; i < mod->smp; i++) {
-			libxmp_free_sample(&m->xsmp[i]);
-		}
-		free(m->xsmp);
-		m->xsmp = NULL;
-	}
-#endif
+	m->midi = NULL;
 
 	libxmp_free_scan(ctx);
 
