@@ -1,5 +1,5 @@
 /* Extended Module Player
- * Copyright (C) 1996-2021 Claudio Matsuoka and Hipolito Carraro Jr
+ * Copyright (C) 1996-2023 Claudio Matsuoka and Hipolito Carraro Jr
  *
  * Permission is hereby granted, free of charge, to any person obtaining a
  * copy of this software and associated documentation files (the "Software"),
@@ -53,12 +53,15 @@ const struct module_quirk mq[] = {
 		XMP_FLAGS_FIXLOOP, XMP_MODE_AUTO
 	},
 
+#if 0
 	/* "siedler ii" (added by Daniel Åkerud) */
+	/* Timing fixed by vblank scan compare. CIA: 32m10s  VBlank: 12m32s */
 	{
 		{ 0x70, 0xaa, 0x03, 0x4d, 0xfb, 0x2f, 0x1f, 0x73,
 		  0xd9, 0xfd, 0xba, 0xfe, 0x13, 0x1b, 0xb7, 0x01 },
 		XMP_FLAGS_VBLANK, XMP_MODE_AUTO
 	},
+#endif
 
 	/* "Klisje paa klisje" (added by Kjetil Torgrim Homme) */
 	{
@@ -99,6 +102,126 @@ const struct module_quirk mq[] = {
 		0, XMP_MODE_PROTRACKER
 	},
 
+	/* grooving3.mod */
+	/* length 150778 crc32c 0xfdcf9aadU */
+	{
+		{ 0xdb, 0x61, 0x22, 0x44, 0x39, 0x85, 0x74, 0xe9,
+		  0xfa, 0x11, 0xb8, 0xfb, 0x87, 0xe8, 0xde, 0xc5, },
+		XMP_FLAGS_VBLANK, XMP_MODE_AUTO
+	},
+	/* mod.Rundgren */
+	/* length 195078 crc32c 0x8fa827a4U */
+	{
+		{ 0x9a, 0xdb, 0xb2, 0x09, 0x07, 0x1c, 0x44, 0x82,
+		  0xc5, 0xdf, 0x83, 0x52, 0xcc, 0x73, 0x9f, 0x20, },
+		XMP_FLAGS_VBLANK, XMP_MODE_AUTO
+	},
+	/* dance feeling by Audiomonster */
+	/* length 169734 crc32c 0x79fa2c9bU */
+	{
+		{ 0x31, 0x2c, 0x3d, 0xaa, 0x5f, 0x1a, 0x54, 0x44,
+		  0x9d, 0xf7, 0xc4, 0x41, 0x8a, 0xc5, 0x01, 0x02, },
+		XMP_FLAGS_VBLANK, XMP_MODE_AUTO
+	},
+	/* knights melody by Audiomonster */
+	/* length 77798 crc32c 0x7bf19c5bU */
+	{
+		{ 0x31, 0xc3, 0x0e, 0x32, 0xfc, 0x99, 0x95, 0xd2,
+		  0x97, 0x20, 0xb3, 0x77, 0x50, 0x05, 0xfe, 0xa5, },
+		XMP_FLAGS_VBLANK, XMP_MODE_AUTO
+	},
+	/* hcomme by Bouffon */
+	/* length 71346 crc32c 0x4ad49cb3U */
+	{
+		{ 0x6e, 0xf9, 0x78, 0xc1, 0x80, 0xae, 0x51, 0x06,
+		  0x05, 0x7c, 0x6e, 0xd0, 0x26, 0x7e, 0xfe, 0x3d, },
+		XMP_FLAGS_VBLANK, XMP_MODE_AUTO
+	},
+	/* ((((aquapool)))) by Dolphin */
+	/* length 62932 crc32c 0x05b103fcU */
+	{
+		{ 0xff, 0x0b, 0xe0, 0x26, 0xc6, 0x31, 0xb5, 0x9b,
+		  0x94, 0x83, 0x94, 0x99, 0x7e, 0x24, 0x7c, 0xdd, },
+		XMP_FLAGS_VBLANK, XMP_MODE_AUTO
+	},
+	/* 100yarddash by Dr. Awesome */
+	/* length 104666 crc32c 0xd2b0e4a6U */
+	{
+		{ 0x5b, 0xff, 0x2f, 0xb8, 0xef, 0x3c, 0xbe, 0x55,
+		  0xa8, 0xe2, 0xa7, 0xcf, 0x5c, 0xbd, 0xdd, 0xb2, },
+		XMP_FLAGS_VBLANK, XMP_MODE_AUTO
+	},
+	/* jazz-reggae-funk by Droid */
+	/* length 115564 crc32c 0x41ff635fU */
+	{
+		{ 0xe5, 0x6e, 0x31, 0x2f, 0x62, 0x80, 0xc1, 0x9d,
+		  0x2f, 0x24, 0x54, 0xf3, 0x89, 0x3f, 0x94, 0x6c, },
+		XMP_FLAGS_VBLANK, XMP_MODE_AUTO
+	},
+	/* hard and heavy by Fish */
+	/* length 69814 crc32c 0x1f09d3d5U */
+	{
+		{ 0x6b, 0xce, 0x39, 0x94, 0x75, 0x42, 0x06, 0x74,
+		  0xd2, 0x83, 0xbc, 0x5e, 0x7b, 0x42, 0x1f, 0xa0, },
+		XMP_FLAGS_VBLANK, XMP_MODE_AUTO
+	},
+	/* crazy valley by Julius and Droid */
+	/* length 97496 crc32c 0xb8eec40eU */
+	{
+		{ 0x23, 0x77, 0x18, 0x1d, 0x21, 0x9b, 0x41, 0x8f,
+		  0xc1, 0xb4, 0xf4, 0xf8, 0x22, 0xdd, 0xd8, 0xb6, },
+		XMP_FLAGS_VBLANK, XMP_MODE_AUTO
+	},
+	/* THE ILLOGICAL ONE by Rhino */
+	/* length 173432 crc32c 0xcb4e2987U */
+	{
+		{ 0xd8, 0xc2, 0xbb, 0xe6, 0x11, 0xd0, 0x5c, 0x02,
+		  0x8e, 0x3b, 0xcb, 0x7c, 0x4a, 0x7d, 0x43, 0xa0, },
+		XMP_FLAGS_VBLANK, XMP_MODE_AUTO
+	},
+	/* sounds of holiday by Spacebrain */
+	/* length 309520 crc32c 0x28804a57U */
+	{
+		{ 0x36, 0x18, 0x19, 0xa4, 0x9d, 0xa2, 0xa2, 0x6f,
+		  0x58, 0x60, 0xc4, 0xd9, 0x0d, 0xa2, 0x9f, 0x49, },
+		XMP_FLAGS_VBLANK, XMP_MODE_AUTO
+	},
+	/* sunisinus by Speed-Head */
+	/* length 175706 crc32c 0x2e56451bU */
+	{
+		{ 0x7e, 0x69, 0x44, 0xb6, 0x38, 0x0d, 0x27, 0x14,
+		  0x70, 0x5d, 0x44, 0xce, 0xce, 0xdd, 0x37, 0x31, },
+		XMP_FLAGS_VBLANK, XMP_MODE_AUTO
+	},
+	/* eat the fulcrum bop by The Assassin */
+	/* length 160286 crc32c 0x583a4683U */
+	{
+		{ 0x11, 0xe9, 0x6f, 0x62, 0xe1, 0xc3, 0xc5, 0xcc,
+		  0x3b, 0xaf, 0xea, 0x69, 0x4b, 0xce, 0x5f, 0xec, },
+		XMP_FLAGS_VBLANK, XMP_MODE_AUTO
+	},
+	/* obvious disaster by Tip */
+	/* length 221086 crc32c 0x51c6d489U */
+	{
+		{ 0x06, 0x8e, 0x69, 0x01, 0x49, 0x8f, 0xbd, 0x0f,
+		  0xfc, 0xb7, 0x8f, 0x2a, 0x91, 0xe1, 0x8b, 0xe8, },
+		XMP_FLAGS_VBLANK, XMP_MODE_AUTO
+	},
+	/* alien nation by Turtle */
+	/* length 167548 crc32c 0xc9ec1674U */
+	{
+		{ 0x71, 0xdf, 0x11, 0xac, 0x5d, 0xec, 0x07, 0xf8,
+		  0x10, 0x6f, 0x28, 0x8d, 0x47, 0x59, 0x54, 0x9b, },
+		XMP_FLAGS_VBLANK, XMP_MODE_AUTO
+	},
+	/* illusions!2 by Zuhl */
+	/* length 289770 crc32c 0x6bf5fbcfU */
+	{
+		{ 0xca, 0x37, 0x8c, 0x0e, 0x87, 0x4f, 0x1e, 0xcd,
+		  0xa3, 0xe9, 0x8b, 0xdd, 0x11, 0x46, 0x8d, 0x69, },
+		XMP_FLAGS_VBLANK, XMP_MODE_AUTO
+	},
+
 	{
 		{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
 		  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
@@ -127,7 +250,7 @@ char *libxmp_adjust_string(char *s)
 	int i;
 
 	for (i = 0; i < strlen(s); i++) {
-		if (!isprint((int)s[i]) || ((uint8) s[i] > 127))
+		if (!isprint((unsigned char)s[i]) || ((uint8) s[i] > 127))
 			s[i] = ' ';
 	}
 
@@ -150,9 +273,21 @@ static void check_envelope(struct xmp_envelope *env)
 		env->flg &= ~XMP_ENVELOPE_LOOP;
 	}
 
-	/* Disable envelope loop if invalid sustain */
-	if (env->sus >= env->npt) {
-		env->flg &= ~XMP_ENVELOPE_ON;
+	/* Disable envelope sustain if invalid sustain */
+	if (env->sus >= env->npt || env->sue >= env->npt) {
+		env->flg &= ~XMP_ENVELOPE_SUS;
+	}
+}
+
+static void clamp_volume_envelope(struct module_data *m, struct xmp_envelope *env)
+{
+	/* Clamp broken values in the volume envelope to the expected range. */
+	if (env->flg & XMP_ENVELOPE_ON) {
+		int i;
+		for (i = 0; i < env->npt; i++) {
+			int16 *data = &env->data[i * 2 + 1];
+			CLAMP(*data, 0, m->volbase);
+		}
 	}
 }
 
@@ -171,26 +306,25 @@ void libxmp_load_prologue(struct context_data *ctx)
 	m->quirk = 0;
 	m->read_event_type = READ_EVENT_MOD;
 	m->period_type = PERIOD_AMIGA;
+	m->compare_vblank = 0;
 	m->comment = NULL;
 	m->scan_cnt = NULL;
+	m->midi = NULL;
 
 	/* Set defaults */
-    	m->mod.pat = 0;
-    	m->mod.trk = 0;
-    	m->mod.chn = 4;
-    	m->mod.ins = 0;
-    	m->mod.smp = 0;
-    	m->mod.spd = 6;
-    	m->mod.bpm = 125;
-    	m->mod.len = 0;
-    	m->mod.rst = 0;
+	m->mod.pat = 0;
+	m->mod.trk = 0;
+	m->mod.chn = 4;
+	m->mod.ins = 0;
+	m->mod.smp = 0;
+	m->mod.spd = 6;
+	m->mod.bpm = 125;
+	m->mod.len = 0;
+	m->mod.rst = 0;
 
 #ifndef LIBXMP_CORE_PLAYER
 	m->extra = NULL;
 #endif
-#ifndef LIBXMP_CORE_DISABLE_IT
-	m->xsmp = NULL;
-#endif
 
 	m->time_factor = DEFAULT_TIME_FACTOR;
 
@@ -209,7 +343,7 @@ void libxmp_load_epilogue(struct context_data *ctx)
 	struct xmp_module *mod = &m->mod;
 	int i, j;
 
-    	mod->gvl = m->gvol;
+	mod->gvl = m->gvol;
 
 	/* Sanity check for module parameters */
 	CLAMP(mod->len, 0, XMP_MAX_MOD_LENGTH);
@@ -230,7 +364,7 @@ void libxmp_load_epilogue(struct context_data *ctx)
 	if (mod->spd <= 0 || mod->spd > 255) {
 		mod->spd = 6;
 	}
-	CLAMP(mod->bpm, XMP_MIN_BPM, 255);
+	CLAMP(mod->bpm, XMP_MIN_BPM, 1000);
 
 	/* Set appropriate values for instrument volumes and subinstrument
 	 * global volumes when QUIRK_INSVOL is not set, to keep volume values
@@ -254,8 +388,28 @@ void libxmp_load_epilogue(struct context_data *ctx)
 		check_envelope(&mod->xxi[i].aei);
 		check_envelope(&mod->xxi[i].fei);
 		check_envelope(&mod->xxi[i].pei);
+		clamp_volume_envelope(m, &mod->xxi[i].aei);
 	}
 
+#ifndef LIBXMP_CORE_DISABLE_IT
+	/* TODO: there's no unintrusive and clean way to get this struct into
+	 * libxmp_load_sample currently, so bound these fields here for now. */
+	for (i = 0; i < mod->smp; i++) {
+		struct xmp_sample *xxs = &mod->xxs[i];
+		struct extra_sample_data *xtra = &m->xtra[i];
+		if (xtra->sus < 0) {
+			xtra->sus = 0;
+		}
+		if (xtra->sue > xxs->len) {
+			xtra->sue = xxs->len;
+		}
+		if (xtra->sus >= xxs->len || xtra->sus >= xtra->sue) {
+			xtra->sus = xtra->sue = 0;
+			xxs->flg &= ~(XMP_SAMPLE_SLOOP | XMP_SAMPLE_SLOOP_BIDIR);
+		}
+	}
+#endif
+
 	p->filter = 0;
 	p->mode = XMP_MODE_AUTO;
 	p->flags = p->player_flags;
@@ -283,7 +437,7 @@ int libxmp_prepare_scan(struct context_data *ctx)
 		return 0;
 	}
 
-	m->scan_cnt = calloc(sizeof (uint8 *), mod->len);
+	m->scan_cnt = (uint8 **) calloc(mod->len, sizeof(uint8 *));
 	if (m->scan_cnt == NULL)
 		return -XMP_ERROR_SYSTEM;
 
@@ -299,7 +453,7 @@ int libxmp_prepare_scan(struct context_data *ctx)
 		}
 
 		pat = pat_idx >= mod->pat ? NULL : mod->xxp[pat_idx];
-		m->scan_cnt[i] = calloc(1, pat && pat->rows ? pat->rows : 1);
+		m->scan_cnt[i] = (uint8 *) calloc(1, (pat && pat->rows)? pat->rows : 1);
 		if (m->scan_cnt[i] == NULL)
 			return -XMP_ERROR_SYSTEM;
 	}
@@ -309,6 +463,7 @@ int libxmp_prepare_scan(struct context_data *ctx)
 
 void libxmp_free_scan(struct context_data *ctx)
 {
+	struct player_data *p = &ctx->p;
 	struct module_data *m = &ctx->m;
 	struct xmp_module *mod = &m->mod;
 	int i;
@@ -320,6 +475,9 @@ void libxmp_free_scan(struct context_data *ctx)
 		free(m->scan_cnt);
 		m->scan_cnt = NULL;
 	}
+
+	free(p->scan);
+	p->scan = NULL;
 }
 
 /* Process player personality flags */
@@ -394,6 +552,9 @@ int libxmp_set_player_mode(struct context_data *ctx)
 		return -1;
 	}
 
+	if (p->mode != XMP_MODE_AUTO)
+		m->compare_vblank = 0;
+
 	return 0;
 }
 
