@@ -1,5 +1,5 @@
 /* Extended Module Player
- * Copyright (C) 1996-2021 Claudio Matsuoka and Hipolito Carraro Jr
+ * Copyright (C) 1996-2022 Claudio Matsuoka and Hipolito Carraro Jr
  *
  * Permission is hereby granted, free of charge, to any person obtaining a
  * copy of this software and associated documentation files (the "Software"),
@@ -22,7 +22,7 @@
 
 #include "loader.h"
 #include "iff.h"
-#include "period.h"
+#include "../period.h"
 
 #define MAGIC_D_T_	MAGIC4('D','.','T','.')
 
@@ -73,7 +73,7 @@ static int get_d_t_(struct module_data *m, int size, HIO_HANDLE *f, void *parm)
 	hio_read16b(f);			/* reserved */
 	mod->spd = hio_read16b(f);
 	if ((b = hio_read16b(f)) > 0)	/* RAMBO.DTM has bpm 0 */
-		mod->bpm = b;
+		mod->bpm = b;		/* Not clamped by Digital Tracker. */
 	hio_read32b(f);			/* undocumented */
 
 	hio_read(mod->name, 32, 1, f);
@@ -160,7 +160,8 @@ static int get_inst(struct module_data *m, int size, HIO_HANDLE *f, void *parm)
 		return -1;
 
 	for (i = 0; i < mod->ins; i++) {
-		int fine, replen, flag;
+		uint32 repstart, replen;
+		int fine, flag;
 
 		if (libxmp_alloc_subinstrument(mod, i, 1) < 0)
 			return -1;
@@ -171,12 +172,15 @@ static int get_inst(struct module_data *m, int size, HIO_HANDLE *f, void *parm)
 		fine = hio_read8s(f);	/* finetune */
 		mod->xxi[i].sub[0].vol = hio_read8(f);
 		mod->xxi[i].sub[0].pan = 0x80;
-		mod->xxs[i].lps = hio_read32b(f);
+		repstart = hio_read32b(f);
 		replen = hio_read32b(f);
-		mod->xxs[i].lpe = mod->xxs[i].lps + replen - 1;
+		mod->xxs[i].lps = repstart;
+		mod->xxs[i].lpe = repstart + replen - 1;
 		mod->xxs[i].flg = replen > 2 ?  XMP_SAMPLE_LOOP : 0;
 
-		hio_read(name, 22, 1, f);
+		if (hio_read(name, 22, 1, f) == 0)
+			return -1;
+
 		libxmp_instrument_name(mod, i, name, 22);
 
 		flag = hio_read16b(f);	/* bit 0-7:resol 8:stereo */
@@ -200,7 +204,7 @@ static int get_inst(struct module_data *m, int size, HIO_HANDLE *f, void *parm)
 			i, mod->xxi[i].name,
 			mod->xxs[i].len,
 			mod->xxs[i].flg & XMP_SAMPLE_16BIT ? '+' : ' ',
-			mod->xxs[i].lps,
+			repstart,
 			replen,
 			mod->xxs[i].flg & XMP_SAMPLE_LOOP ? 'L' : ' ',
 			flag & 0x100 ? 'S' : ' ',
