@@ -20,6 +20,11 @@
  * THE SOFTWARE.
  */
 
+#ifndef LIBXMP_LOADERS_IT_H
+#define LIBXMP_LOADERS_IT_H
+
+#include "loader.h"
+
 /* IT flags */
 #define IT_STEREO	0x01
 #define IT_VOL_OPT	0x02	/* Not recognized */
@@ -27,9 +32,14 @@
 #define IT_LINEAR_FREQ	0x08
 #define IT_OLD_FX	0x10
 #define IT_LINK_GXX	0x20
+#define IT_MIDI_WHEEL	0x40
+#define IT_MIDI_CONFIG	0x80
 
 /* IT special */
 #define IT_HAS_MSG	0x01
+#define IT_EDIT_HISTORY	0x02
+#define IT_HIGHLIGHTS	0x04
+#define IT_SPEC_MIDICFG	0x08
 
 /* IT instrument flags */
 #define IT_INST_SAMPLE	0x01
@@ -56,6 +66,7 @@
 #define IT_CVT_DIFF	0x04	/* Compressed sample flag */
 #define IT_CVT_BYTEDIFF	0x08	/* 'safe to ignore' according to ittech.txt */
 #define IT_CVT_12BIT	0x10	/* 'safe to ignore' according to ittech.txt */
+#define IT_CVT_ADPCM	0xff	/* Special: always indicates Modplug ADPCM4 */
 
 /* IT envelope flags */
 #define IT_ENV_ON	0x01
@@ -179,3 +190,7 @@ struct it_sample_header {
 	uint8 vit;		/* Vibrato waveform */
 };
 
+int itsex_decompress8(HIO_HANDLE *src, uint8 *dst, int len, int it215);
+int itsex_decompress16(HIO_HANDLE *src, int16 *dst, int len, int it215);
+
+#endif /* LIBXMP_LOADERS_IT_H */
