@@ -1,5 +1,5 @@
 /* Extended Module Player
- * Copyright (C) 1996-2021 Claudio Matsuoka and Hipolito Carraro Jr
+ * Copyright (C) 1996-2023 Claudio Matsuoka and Hipolito Carraro Jr
  *
  * Permission is hereby granted, free of charge, to any person obtaining a
  * copy of this software and associated documentation files (the "Software"),
@@ -24,13 +24,12 @@
 
 #include "loader.h"
 #include "it.h"
-#include "period.h"
+#include "../period.h"
 
 #define MAGIC_IMPM	MAGIC4('I','M','P','M')
 #define MAGIC_IMPI	MAGIC4('I','M','P','I')
 #define MAGIC_IMPS	MAGIC4('I','M','P','S')
 
-
 static int it_test(HIO_HANDLE *, char *, const int);
 static int it_load(struct module_data *, HIO_HANDLE *, const int);
 
@@ -55,7 +54,6 @@ static int it_test(HIO_HANDLE *f, char *t, const int start)
 #define FX_XTND 0xfe
 #define L_CHANNELS 64
 
-
 static const uint8 fx[32] = {
 	/*   */ FX_NONE,
 	/* A */ FX_S3M_SPEED,
@@ -83,19 +81,14 @@ static const uint8 fx[32] = {
 	/* W */ FX_GVOL_SLIDE,
 	/* X */ FX_SETPAN,
 	/* Y */ FX_PANBRELLO,
-	/* Z */ FX_FLT_CUTOFF,
-	/* ? */ FX_NONE,
+	/* Z */ FX_MACRO,
 	/* ? */ FX_NONE,
+	/* / */ FX_MACROSMOOTH,
 	/* ? */ FX_NONE,
 	/* ? */ FX_NONE,
 	/* ? */ FX_NONE
 };
 
-
-int itsex_decompress8 (HIO_HANDLE *, void *, int, int);
-int itsex_decompress16 (HIO_HANDLE *, void *, int, int);
-
-
 static void xlat_fx(int c, struct xmp_event *e, uint8 *last_fxp, int new_fx)
 {
 	uint8 h = MSN(e->fxp), l = LSN(e->fxp);
@@ -144,9 +137,16 @@ static void xlat_fx(int c, struct xmp_event *e, uint8 *last_fxp, int new_fx)
 			e->fxt = FX_SETPAN;
 			e->fxp = l << 4;
 			break;
-		case 0x9:	/* 0x91 = set surround */
-			e->fxt = FX_SURROUND;
-			e->fxp = l;
+		case 0x9:
+			if (l == 0 || l == 1) {
+				/* 0x91 = set surround */
+				e->fxt = FX_SURROUND;
+				e->fxp = l;
+			} else if (l == 0xe || l == 0xf) {
+				/* 0x9f Play reverse (MPT) */
+				e->fxt = FX_REVERSE;
+				e->fxp = l - 0xe;
+			}
 			break;
 		case 0xa:	/* High offset */
 			e->fxt = FX_HIOFFSET;
@@ -165,18 +165,14 @@ static void xlat_fx(int c, struct xmp_event *e, uint8 *last_fxp, int new_fx)
 			e->fxt = FX_IT_ROWDELAY;
 			e->fxp = l;
 			break;
+		case 0xf:	/* Set parametered macro */
+			e->fxt = FX_MACRO_SET;
+			e->fxp = l;
+			break;
 		default:
 			e->fxt = e->fxp = 0;
 		}
 		break;
-	case FX_FLT_CUTOFF:
-		if (e->fxp > 0x7f && e->fxp < 0x90) {	/* Resonance */
-			e->fxt = FX_FLT_RESN;
-			e->fxp = (e->fxp - 0x80) * 16;
-		} else {	/* Cutoff */
-			e->fxp *= 2;
-		}
-		break;
 	case FX_TREMOR:
 		if (!new_fx && e->fxp != 0) {
 			e->fxp = ((MSN(e->fxp) + 1) << 4) | (LSN(e->fxp) + 1);
@@ -258,6 +254,34 @@ static void fix_name(uint8 *s, int l)
 }
 
 
+static int load_it_midi_config(struct module_data *m, HIO_HANDLE *f)
+{
+	int i;
+
+	m->midi = (struct midi_macro_data *) calloc(1, sizeof(struct midi_macro_data));
+	if (m->midi == NULL)
+		return -1;
+
+	/* Skip global MIDI macros */
+	if (hio_seek(f, 9 * 32, SEEK_CUR) < 0)
+		return -1;
+
+	/* SFx macros */
+	for (i = 0; i < 16; i++) {
+		if (hio_read(m->midi->param[i].data, 1, 32, f) < 32)
+			return -1;
+		m->midi->param[i].data[31] = '\0';
+	}
+	/* Zxx macros */
+	for (i = 0; i < 128; i++) {
+		if (hio_read(m->midi->fixed[i].data, 1, 32, f) < 32)
+			return -1;
+		m->midi->fixed[i].data[31] = '\0';
+	}
+	return 0;
+}
+
+
 static int read_envelope(struct xmp_envelope *ei, struct it_envelope *env,
 			  HIO_HANDLE *f)
 {
@@ -313,7 +337,8 @@ static int read_envelope(struct xmp_envelope *ei, struct it_envelope *env,
 	return 0;
 }
 
-static void identify_tracker(struct module_data *m, struct it_file_header *ifh)
+static void identify_tracker(struct module_data *m, struct it_file_header *ifh,
+			     int pat_before_smp, int *is_mpt_116)
 {
 #ifndef LIBXMP_CORE_PLAYER
 	char tracker_name[40];
@@ -337,6 +362,16 @@ static void identify_tracker(struct module_data *m, struct it_file_header *ifh)
 			strcpy(tracker_name, "ModPlug Tracker 1.16");
 			/* ModPlug Tracker files aren't really IMPM 2.00 */
 			ifh->cmwt = sample_mode ? 0x100 : 0x214;
+			*is_mpt_116 = 1;
+		} else if (ifh->cmwt == 0x0200 && ifh->cwt == 0x0202 && pat_before_smp) {
+			/* ModPlug Tracker ITs from pre-alpha 4 use tracker
+			 * 0x0202 and format 0x0200. Unfortunately, ITs from
+			 * Impulse Tracker may *also* use this. These MPT ITs
+			 * can be detected because they write patterns before
+			 * samples/instruments. */
+			strcpy(tracker_name, "ModPlug Tracker 1.0 pre-alpha");
+			ifh->cmwt = sample_mode ? 0x100 : 0x200;
+			*is_mpt_116 = 1;
 		} else if (ifh->cwt == 0x0216) {
 			strcpy(tracker_name, "Impulse Tracker 2.14v3");
 		} else if (ifh->cwt == 0x0217) {
@@ -352,6 +387,7 @@ static void identify_tracker(struct module_data *m, struct it_file_header *ifh)
 	case 0x7f:
 		if (ifh->cwt == 0x0888) {
 			strcpy(tracker_name, "OpenMPT 1.17");
+			*is_mpt_116 = 1;
 		} else if (ifh->cwt == 0x7fff) {
 			strcpy(tracker_name, "munch.py");
 		} else {
@@ -486,7 +522,7 @@ static int load_old_it_instrument(struct xmp_instrument *xxi, HIO_HANDLE *f)
 	xxi->vol = 0x40;
 
 	if (k) {
-		xxi->sub = calloc(sizeof(struct xmp_subinstrument), k);
+		xxi->sub = (struct xmp_subinstrument *) calloc(k, sizeof(struct xmp_subinstrument));
 		if (xxi->sub == NULL) {
 			return -1;
 		}
@@ -637,7 +673,7 @@ static int load_new_it_instrument(struct xmp_instrument *xxi, HIO_HANDLE *f)
 	xxi->vol = i2h.gbv >> 1;
 
 	if (k) {
-		xxi->sub = calloc(sizeof(struct xmp_subinstrument), k);
+		xxi->sub = (struct xmp_subinstrument *) calloc(k, sizeof(struct xmp_subinstrument));
 		if (xxi->sub == NULL)
 			return -1;
 
@@ -671,7 +707,7 @@ static int load_new_it_instrument(struct xmp_instrument *xxi, HIO_HANDLE *f)
 	return 0;
 }
 
-static void force_sample_length(struct xmp_sample *xxs, int len)
+static void force_sample_length(struct xmp_sample *xxs, struct extra_sample_data *xtra, int len)
 {
 	xxs->len = len;
 
@@ -680,6 +716,14 @@ static void force_sample_length(struct xmp_sample *xxs, int len)
 
 	if (xxs->lps >= xxs->len)
 		xxs->flg &= ~XMP_SAMPLE_LOOP;
+
+	if (xtra) {
+		if (xtra->sue > xxs->len)
+			xtra->sue = xxs->len;
+
+		if(xtra->sus >= xxs->len)
+			xxs->flg &= ~(XMP_SAMPLE_SLOOP | XMP_SAMPLE_SLOOP_BIDIR);
+	}
 }
 
 static int load_it_sample(struct module_data *m, int i, int start,
@@ -687,12 +731,13 @@ static int load_it_sample(struct module_data *m, int i, int start,
 {
 	struct it_sample_header ish;
 	struct xmp_module *mod = &m->mod;
-	struct xmp_sample *xxs, *xsmp;
+	struct extra_sample_data *xtra;
+	struct xmp_sample *xxs;
 	int j, k;
 	uint8 buf[80];
 
 	if (sample_mode) {
-		mod->xxi[i].sub = calloc(sizeof(struct xmp_subinstrument), 1);
+		mod->xxi[i].sub = (struct xmp_subinstrument *) calloc(1, sizeof(struct xmp_subinstrument));
 		if (mod->xxi[i].sub == NULL) {
 			return -1;
 		}
@@ -711,7 +756,7 @@ static int load_it_sample(struct module_data *m, int i, int start,
 	}
 
 	xxs = &mod->xxs[i];
-	xsmp = &m->xsmp[i];
+	xtra = &m->xtra[i];
 
 	memcpy(ish.dosname, buf + 4, 12);
 	ish.zero = buf[16];
@@ -749,14 +794,8 @@ static int load_it_sample(struct module_data *m, int i, int start,
 	xxs->flg |= ish.flags & IT_SMP_BSLOOP ? XMP_SAMPLE_SLOOP_BIDIR : 0;
 
 	if (ish.flags & IT_SMP_SLOOP) {
-		memcpy(xsmp, xxs, sizeof (struct xmp_sample));
-		xsmp->lps = ish.sloopbeg;
-		xsmp->lpe = ish.sloopend;
-		xsmp->flg |= XMP_SAMPLE_LOOP;
-		xsmp->flg &= ~XMP_SAMPLE_LOOP_BIDIR;
-		if (ish.flags & IT_SMP_BSLOOP) {
-			xsmp->flg |= XMP_SAMPLE_LOOP_BIDIR;
-		}
+		xtra->sus = ish.sloopbeg;
+		xtra->sue = ish.sloopend;
 	}
 
 	if (sample_mode) {
@@ -827,13 +866,16 @@ static int load_it_sample(struct module_data *m, int i, int start,
 		if (xxs->lpe > xxs->len || xxs->lps >= xxs->lpe)
 			xxs->flg &= ~XMP_SAMPLE_LOOP;
 
+		if (ish.convert == IT_CVT_ADPCM)
+			cvt |= SAMPLE_FLAG_ADPCM;
+
 		if (~ish.convert & IT_CVT_SIGNED)
 			cvt |= SAMPLE_FLAG_UNS;
 
 		/* compressed samples */
 		if (ish.flags & IT_SMP_COMP) {
 			long min_size, file_len, left;
-			uint8 *buf;
+			void *decbuf;
 			int ret;
 
 			/* Sanity check - the lower bound on IT compressed
@@ -853,17 +895,15 @@ static int load_it_sample(struct module_data *m, int i, int start,
 				   "resizing to %ld",
 				   i, xxs->len, min_size, left, left << 3);
 
-				force_sample_length(xxs, left << 3);
-				if (ish.flags & IT_SMP_SLOOP)
-					force_sample_length(xsmp, left << 3);
+				force_sample_length(xxs, xtra, left << 3);
 			}
 
-			buf = calloc(1, xxs->len * 2);
-			if (buf == NULL)
+			decbuf = (uint8 *) calloc(1, xxs->len * 2);
+			if (decbuf == NULL)
 				return -1;
 
 			if (ish.flags & IT_SMP_16BIT) {
-				itsex_decompress16(f, buf, xxs->len,
+				itsex_decompress16(f, (int16 *)decbuf, xxs->len,
 						   ish.convert & IT_CVT_DIFF);
 
 #ifdef WORDS_BIGENDIAN
@@ -873,44 +913,19 @@ static int load_it_sample(struct module_data *m, int i, int start,
 				cvt |= SAMPLE_FLAG_BIGEND;
 #endif
 			} else {
-				itsex_decompress8(f, buf, xxs->len,
+				itsex_decompress8(f, (uint8 *)decbuf, xxs->len,
 						  ish.convert & IT_CVT_DIFF);
 			}
 
-			if (ish.flags & IT_SMP_SLOOP) {
-				long pos = hio_tell(f);
-				if (pos < 0) {
-					free(buf);
-					return -1;
-				}
-				ret = libxmp_load_sample(m, NULL, SAMPLE_FLAG_NOLOAD |
-							cvt, &m->xsmp[i], buf);
-				if (ret < 0) {
-					free(buf);
-					return -1;
-				}
-				hio_seek(f, pos, SEEK_SET);
-			}
-
 			ret = libxmp_load_sample(m, NULL, SAMPLE_FLAG_NOLOAD | cvt,
-					  &mod->xxs[i], buf);
+					  &mod->xxs[i], decbuf);
 			if (ret < 0) {
-				free(buf);
+				free(decbuf);
 				return -1;
 			}
 
-			free(buf);
+			free(decbuf);
 		} else {
-			if (ish.flags & IT_SMP_SLOOP) {
-				long pos = hio_tell(f);
-				if (pos < 0) {
-					return -1;
-				}
-				if (libxmp_load_sample(m, f, cvt, &m->xsmp[i], NULL) < 0)
-					return -1;
-				hio_seek(f, pos, SEEK_SET);
-			}
-
 			if (libxmp_load_sample(m, f, cvt, &mod->xxs[i], NULL) < 0)
 				return -1;
 		}
@@ -927,7 +942,7 @@ static int load_it_pattern(struct module_data *m, int i, int new_fx,
 	uint8 mask[L_CHANNELS];
 	uint8 last_fxp[64];
 
-	int r, c, pat_len;
+	int r, c, pat_len, num_rows;
 	uint8 b;
 
 	r = 0;
@@ -937,7 +952,7 @@ static int load_it_pattern(struct module_data *m, int i, int new_fx,
 	memset(&dummy, 0, sizeof(struct xmp_event));
 
 	pat_len = hio_read16l(f) /* - 4 */ ;
-	mod->xxp[i]->rows = hio_read16l(f);
+	mod->xxp[i]->rows = num_rows = hio_read16l(f);
 
 	if (libxmp_alloc_tracks_in_pattern(mod, i) < 0) {
 		return -1;
@@ -947,7 +962,7 @@ static int load_it_pattern(struct module_data *m, int i, int new_fx,
 	hio_read16l(f);
 	hio_read16l(f);
 
-	while (--pat_len >= 0) {
+	while (r < num_rows && --pat_len >= 0) {
 		b = hio_read8(f);
 		if (hio_error(f)) {
 			return -1;
@@ -968,7 +983,7 @@ static int load_it_pattern(struct module_data *m, int i, int new_fx,
 		 * real number of channels before loading the patterns and
 		 * we don't want to set it to 64 channels.
 		 */
-		if (c >= mod->chn || r >= mod->xxp[i]->rows) {
+		if (c >= mod->chn) {
 			event = &dummy;
 		} else {
 			event = &EVENT(i, c, r);
@@ -1056,6 +1071,8 @@ static int it_load(struct module_data *m, HIO_HANDLE *f, const int start)
 	uint32 *pp_smp;		/* Pointers to samples */
 	uint32 *pp_pat;		/* Pointers to patterns */
 	int new_fx, sample_mode;
+	int pat_before_smp = 0;
+	int is_mpt_116 = 0;
 
 	LOAD_INIT();
 
@@ -1087,8 +1104,8 @@ static int it_load(struct module_data *m, HIO_HANDLE *f, const int start)
 	ifh.pwd = hio_read8(f);
 
 	/* Sanity check */
-	if (ifh.gv > 0x80 || ifh.mv > 0x80) {
-		D_(D_CRIT "invalid gv (%u) or mv (%u)", ifh.gv, ifh.mv);
+	if (ifh.gv > 0x80) {
+		D_(D_CRIT "invalid gv (%u)", ifh.gv);
 		goto err;
 	}
 
@@ -1099,6 +1116,11 @@ static int it_load(struct module_data *m, HIO_HANDLE *f, const int start)
 	hio_read(ifh.chpan, 64, 1, f);
 	hio_read(ifh.chvol, 64, 1, f);
 
+	if (hio_error(f)) {
+		D_(D_CRIT "error reading IT header");
+		goto err;
+	}
+
 	memcpy(mod->name, ifh.name, sizeof(ifh.name));
 	/* sizeof(ifh.name) == 26, sizeof(mod->name) == 64. */
 	mod->name[sizeof(ifh.name)] = '\0';
@@ -1115,18 +1137,18 @@ static int it_load(struct module_data *m, HIO_HANDLE *f, const int start)
 	}
 
 	if (mod->ins) {
-		pp_ins = calloc(4, mod->ins);
+		pp_ins = (uint32 *) calloc(4, mod->ins);
 		if (pp_ins == NULL)
 			goto err;
 	} else {
 		pp_ins = NULL;
 	}
 
-	pp_smp = calloc(4, mod->smp);
+	pp_smp = (uint32 *) calloc(4, mod->smp);
 	if (pp_smp == NULL)
 		goto err2;
 
-	pp_pat = calloc(4, mod->pat);
+	pp_pat = (uint32 *) calloc(4, mod->pat);
 	if (pp_pat == NULL)
 		goto err3;
 
@@ -1178,9 +1200,22 @@ static int it_load(struct module_data *m, HIO_HANDLE *f, const int start)
 	for (i = 0; i < mod->pat; i++)
 		pp_pat[i] = hio_read32l(f);
 
+	if ((ifh.flags & IT_MIDI_CONFIG) || (ifh.special & IT_SPEC_MIDICFG)) {
+		/* Skip edit history if it exists. */
+		if (ifh.special & IT_EDIT_HISTORY) {
+			int skip = hio_read16l(f) * 8;
+			if (hio_error(f) || (skip && hio_seek(f, skip, SEEK_CUR) < 0))
+				goto err4;
+		}
+		if (load_it_midi_config(m, f) < 0)
+			goto err4;
+	}
+	if (mod->smp && mod->pat && pp_pat[0] != 0 && pp_pat[0] < pp_smp[0])
+		pat_before_smp = 1;
+
 	m->c4rate = C4_NTSC_RATE;
 
-	identify_tracker(m, &ifh);
+	identify_tracker(m, &ifh, pat_before_smp, &is_mpt_116);
 
 	MODULE_INFO();
 
@@ -1194,14 +1229,6 @@ static int it_load(struct module_data *m, HIO_HANDLE *f, const int start)
 	if (libxmp_init_instrument(m) < 0)
 		goto err4;
 
-	/* Alloc extra samples for sustain loop */
-	if (mod->smp > 0) {
-		m->xsmp = calloc(sizeof (struct xmp_sample), mod->smp);
-		if (m->xsmp == NULL) {
-			goto err4;
-		}
-	}
-
 	D_(D_INFO "Instruments: %d", mod->ins);
 
 	for (i = 0; i < mod->ins; i++) {
@@ -1258,7 +1285,7 @@ static int it_load(struct module_data *m, HIO_HANDLE *f, const int start)
 	max_ch = 0;
 	for (i = 0; i < mod->pat; i++) {
 		uint8 mask[L_CHANNELS];
-		int pat_len;
+		int pat_len, num_rows, row;
 
 		/* If the offset to a pattern is 0, the pattern is empty */
 		if (pp_pat[i] == 0)
@@ -1266,19 +1293,33 @@ static int it_load(struct module_data *m, HIO_HANDLE *f, const int start)
 
 		hio_seek(f, start + pp_pat[i], SEEK_SET);
 		pat_len = hio_read16l(f) /* - 4 */ ;
-		hio_read16l(f);
+		num_rows = hio_read16l(f);
 		memset(mask, 0, L_CHANNELS);
 		hio_read16l(f);
 		hio_read16l(f);
 
-		while (--pat_len >= 0) {
+		/* Sanity check:
+		 * - Impulse Tracker and Schism Tracker allow up to 200 rows.
+		 * - ModPlug Tracker 1.16 allows 256 rows.
+		 * - OpenMPT allows 1024 rows.
+		 */
+		if (num_rows > 1024) {
+			D_(D_WARN "skipping pattern %d (%d rows)", i, num_rows);
+			pp_pat[i] = 0;
+			continue;
+		}
+
+		row = 0;
+		while (row < num_rows && --pat_len >= 0) {
 			int b = hio_read8(f);
 			if (hio_error(f)) {
 				D_(D_CRIT "error scanning pattern %d", i);
 				goto err4;
 			}
-			if (b == 0)
+			if (b == 0) {
+				row++;
 				continue;
+			}
 
 			c = (b - 1) & 63;
 
@@ -1354,9 +1395,8 @@ static int it_load(struct module_data *m, HIO_HANDLE *f, const int start)
 	free(pp_ins);
 
 	/* Song message */
-
 	if (ifh.special & IT_HAS_MSG) {
-		if ((m->comment = malloc(ifh.msglen)) != NULL) {
+		if ((m->comment = (char *)malloc(ifh.msglen)) != NULL) {
 			hio_seek(f, start + ifh.msgofs, SEEK_SET);
 
 			D_(D_INFO "Message length : %d", ifh.msglen);
@@ -1401,8 +1441,15 @@ static int it_load(struct module_data *m, HIO_HANDLE *f, const int start)
 
 	m->gvolbase = 0x80;
 	m->gvol = ifh.gv;
+	m->mvolbase = 48;
+	m->mvol = ifh.mv;
 	m->read_event_type = READ_EVENT_IT;
 
+#ifndef LIBXMP_CORE_PLAYER
+	if (is_mpt_116)
+		libxmp_apply_mpt_preamp(m);
+#endif
+
 	return 0;
 
 err4:
