@@ -3,16 +3,22 @@
 /* Public domain IT sample decompressor by Olivier Lapicque */
 
 #include "loader.h"
+#include "it.h"
 
-static inline uint32 read_bits(HIO_HANDLE *ibuf, uint32 *bitbuf, int *bitnum, int n)
+static inline uint32 read_bits(HIO_HANDLE *ibuf, uint32 *bitbuf, int *bitnum, int n, int *err)
 {
 	uint32 retval = 0;
 	int i = n;
-	int bnum = *bitnum, bbuf = *bitbuf;
+	int bnum = *bitnum;
+	uint32 bbuf = *bitbuf;
 
-	if (n > 0) {
+	if (n > 0 && n <= 32) {
 		do {
 			if (bnum == 0) {
+				if (hio_eof(ibuf)) {
+					*err = EOF;
+					return 0;
+				}
 				bbuf = hio_read8(ibuf);
 				bnum = 8;
 			}
@@ -27,6 +33,10 @@ static inline uint32 read_bits(HIO_HANDLE *ibuf, uint32 *bitbuf, int *bitnum, in
 
 		*bitnum = bnum;
 		*bitbuf = bbuf;
+	} else {
+		/* Invalid shift value. */
+		*err = -2;
+		return 0;
 	}
 
 	return (retval >> (32 - i));
@@ -41,6 +51,7 @@ int itsex_decompress8(HIO_HANDLE *src, uint8 *dst, int len, int it215)
 	int bitnum = 0;
 	uint8 left = 0, temp = 0, temp2 = 0;
 	uint32 d, pos;
+	int err = 0;
 
 	while (len) {
 		if (!block_count) {
@@ -58,8 +69,8 @@ int itsex_decompress8(HIO_HANDLE *src, uint8 *dst, int len, int it215)
 		/* Unpacking */
 		pos = 0;
 		do {
-			uint16 bits = read_bits(src, &bitbuf, &bitnum, left);
-			if (hio_eof(src))
+			uint16 bits = read_bits(src, &bitbuf, &bitnum, left, &err);
+			if (err != 0)
 				return -1;
 
 			if (left < 7) {
@@ -67,9 +78,9 @@ int itsex_decompress8(HIO_HANDLE *src, uint8 *dst, int len, int it215)
 				uint32 j = bits & 0xffff;
 				if (i != j)
 					goto unpack_byte;
-				bits = (read_bits(src, &bitbuf, &bitnum, 3)
+				bits = (read_bits(src, &bitbuf, &bitnum, 3, &err)
 								+ 1) & 0xff;
-				if (hio_eof(src))
+				if (err != 0)
 					return -1;
 
 				left = ((uint8)bits < left) ?  (uint8)bits :
@@ -137,6 +148,7 @@ int itsex_decompress16(HIO_HANDLE *src, int16 *dst, int len, int it215)
 	uint8 left = 0;
 	int16 temp = 0, temp2 = 0;
 	uint32 d, pos;
+	int err = 0;
 
 	while (len) {
 		if (!block_count) {
@@ -154,8 +166,8 @@ int itsex_decompress16(HIO_HANDLE *src, int16 *dst, int len, int it215)
 		/* Unpacking */
 		pos = 0;
 		do {
-			uint32 bits = read_bits(src, &bitbuf, &bitnum, left);
-			if (hio_eof(src))
+			uint32 bits = read_bits(src, &bitbuf, &bitnum, left, &err);
+			if (err != 0)
 				return -1;
 
 			if (left < 7) {
@@ -165,9 +177,8 @@ int itsex_decompress16(HIO_HANDLE *src, int16 *dst, int len, int it215)
 				if (i != j)
 					goto unpack_byte;
 
-				bits = read_bits(src, &bitbuf, &bitnum, 4) + 1;
-
-				if (hio_eof(src))
+				bits = read_bits(src, &bitbuf, &bitnum, 4, &err) + 1;
+				if (err != 0)
 					return -1;
 
 				left = ((uint8)(bits & 0xff) < left) ?
