@@ -1,10 +1,10 @@
 #ifndef XMP_LOADER_H
 #define XMP_LOADER_H
 
-#include "common.h"
-#include "effects.h"
-#include "format.h"
-#include "hio.h"
+#include "../common.h"
+#include "../effects.h"
+#include "../format.h"
+#include "../hio.h"
 
 /* Sample flags */
 #define SAMPLE_FLAG_DIFF	0x0001	/* Differential */
@@ -20,6 +20,10 @@
 #define SAMPLE_FLAG_HSC		0x2000	/* HSC Adlib synth instrument */
 #define SAMPLE_FLAG_ADPCM	0x4000	/* ADPCM4 encoded samples */
 
+/* libxmp_test_name flags */
+#define TEST_NAME_IGNORE_AFTER_0	0x0001
+#define TEST_NAME_IGNORE_AFTER_CR	0x0002
+
 #define DEFPAN(x) (0x80 + ((x) - 0x80) * m->defpan / 100)
 
 int	libxmp_init_instrument		(struct module_data *);
@@ -30,16 +34,18 @@ int	libxmp_alloc_pattern		(struct xmp_module *, int);
 int	libxmp_alloc_track		(struct xmp_module *, int, int);
 int	libxmp_alloc_tracks_in_pattern	(struct xmp_module *, int);
 int	libxmp_alloc_pattern_tracks	(struct xmp_module *, int, int);
+#ifndef LIBXMP_CORE_PLAYER
 int	libxmp_alloc_pattern_tracks_long(struct xmp_module *, int, int);
+#endif
 char	*libxmp_instrument_name		(struct xmp_module *, int, uint8 *, int);
 
 char	*libxmp_copy_adjust		(char *, uint8 *, int);
 int	libxmp_copy_name_for_fopen	(char *, const char *, int);
-int	libxmp_test_name		(uint8 *, int);
+int	libxmp_test_name		(const uint8 *, int, int);
 void	libxmp_read_title		(HIO_HANDLE *, char *, int);
 void	libxmp_set_xxh_defaults		(struct xmp_module *);
-void	libxmp_decode_protracker_event	(struct xmp_event *, uint8 *);
-void	libxmp_decode_noisetracker_event(struct xmp_event *, uint8 *);
+void	libxmp_decode_protracker_event	(struct xmp_event *, const uint8 *);
+void	libxmp_decode_noisetracker_event(struct xmp_event *, const uint8 *);
 void	libxmp_disable_continue_fx	(struct xmp_event *);
 int	libxmp_check_filename_case	(const char *, const char *, char *, int);
 void	libxmp_get_instrument_path	(struct module_data *, char *, int);
@@ -47,7 +53,10 @@ void	libxmp_set_type			(struct module_data *, const char *, ...);
 int	libxmp_load_sample		(struct module_data *, HIO_HANDLE *, int,
 					 struct xmp_sample *, const void *);
 void	libxmp_free_sample		(struct xmp_sample *);
+#ifndef LIBXMP_CORE_PLAYER
 void	libxmp_schism_tracker_string	(char *, size_t, int, int);
+void	libxmp_apply_mpt_preamp	(struct module_data *m);
+#endif
 
 extern uint8		libxmp_ord_xlat[];
 extern const int	libxmp_arch_vol_table[];
