@@ -1,5 +1,5 @@
 /* Extended Module Player
- * Copyright (C) 1996-2021 Claudio Matsuoka and Hipolito Carraro Jr
+ * Copyright (C) 1996-2023 Claudio Matsuoka and Hipolito Carraro Jr
  *
  * Permission is hereby granted, free of charge, to any person obtaining a
  * copy of this software and associated documentation files (the "Software"),
@@ -26,7 +26,7 @@
 
 #include "med.h"
 #include "loader.h"
-#include "med_extras.h"
+#include "../med_extras.h"
 
 static int mmd1_test(HIO_HANDLE *, char *, const int);
 static int mmd1_load (struct module_data *, HIO_HANDLE *, const int);
@@ -45,7 +45,7 @@ static int mmd1_test(HIO_HANDLE *f, char *t, const int start)
 	if (hio_read(id, 1, 4, f) < 4)
 		return -1;
 
-	if (memcmp(id, "MMD0", 4) && memcmp(id, "MMD1", 4))
+	if (memcmp(id, "MMD0", 4) && memcmp(id, "MMD1", 4) && memcmp(id, "MMDC", 4))
 		return -1;
 
 	hio_seek(f, 28, SEEK_CUR);
@@ -64,8 +64,6 @@ static int mmd1_test(HIO_HANDLE *f, char *t, const int start)
 	return 0;
 }
 
-/* Number of octaves in IFFOCT samples */
-static const int num_oct[6] = { 5, 3, 2, 4, 6, 7 };
 
 static int mmd1_load(struct module_data *m, HIO_HANDLE *f, const int start)
 {
@@ -74,14 +72,15 @@ static int mmd1_load(struct module_data *m, HIO_HANDLE *f, const int start)
 	struct MMD0 header;
 	struct MMD0song song;
 	struct MMD1Block block;
-	struct InstrHdr instr;
-	struct SynthInstr synth;
-	struct InstrExt exp_smp;
+	struct InstrExt *exp_smp = NULL;
 	struct MMD0exp expdata;
 	struct xmp_event *event;
+	uint32 *blockarr = NULL;
+	uint32 *smplarr = NULL;
+	uint8 *patbuf = NULL;
 	int ver = 0;
+	int mmdc = 0;
 	int smp_idx = 0;
-	uint8 e[4];
 	int song_offset;
 	int blockarr_offset;
 	int smplarr_offset;
@@ -90,15 +89,19 @@ static int mmd1_load(struct module_data *m, HIO_HANDLE *f, const int start)
 	int songname_offset;
 	int iinfo_offset;
 	int annotxt_offset;
-	int pos;
 	int bpm_on, bpmlen, med_8ch, hexvol;
-	char name[40];
+	int max_lines;
+	int retval = -1;
 
 	LOAD_INIT();
 
 	hio_read(&header.id, 4, 1, f);
 
 	ver = *((char *)&header.id + 3) - '1' + 1;
+	if (ver > 1) {
+		ver = 0;
+		mmdc = 1;
+	}
 
 	D_(D_WARN "load header");
 	header.modlen = hio_read32b(f);
@@ -127,8 +130,10 @@ static int mmd1_load(struct module_data *m, HIO_HANDLE *f, const int start)
 	 * song structure
 	 */
 	D_(D_WARN "load song");
-	if (hio_seek(f, start + song_offset, SEEK_SET) != 0)
-	  return -1;
+	if (hio_seek(f, start + song_offset, SEEK_SET) != 0) {
+		D_(D_CRIT "seek error at song");
+		return -1;
+	}
 	for (i = 0; i < 63; i++) {
 		song.sample[i].rep = hio_read16b(f);
 		song.sample[i].replen = hio_read16b(f);
@@ -142,6 +147,8 @@ static int mmd1_load(struct module_data *m, HIO_HANDLE *f, const int start)
 
 	/* Sanity check */
 	if (song.numblocks > 255 || song.songlen > 256) {
+		D_(D_CRIT "unsupported block count (%d) or song length (%d)",
+		   song.numblocks, song.songlen);
 		return -1;
 	}
 
@@ -160,6 +167,7 @@ static int mmd1_load(struct module_data *m, HIO_HANDLE *f, const int start)
 
 	/* Sanity check */
 	if (song.numsamples > 63) {
+		D_(D_CRIT "invalid instrument count %d", song.numsamples);
 		return -1;
 	}
 
@@ -176,7 +184,7 @@ static int mmd1_load(struct module_data *m, HIO_HANDLE *f, const int start)
 
 	mmd_set_bpm(m, med_8ch, song.deftempo, bpm_on, bpmlen);
 
-        mod->spd = song.tempo2;
+	mod->spd = song.tempo2;
 	mod->pat = song.numblocks;
 	mod->ins = song.numsamples;
 	mod->len = song.songlen;
@@ -185,34 +193,55 @@ static int mmd1_load(struct module_data *m, HIO_HANDLE *f, const int start)
 	memcpy(mod->xxo, song.playseq, mod->len);
 	mod->name[0] = 0;
 
+	/*
+	 * Read smplarr
+	 */
+	D_(D_WARN "read smplarr");
+	smplarr = (uint32 *) malloc(mod->ins * sizeof(uint32));
+	if (smplarr == NULL) {
+		return -1;
+	}
+	if (hio_seek(f, start + smplarr_offset, SEEK_SET) != 0) {
+		D_(D_CRIT "seek error at smplarr");
+		goto err_cleanup;
+	}
+	for (i = 0; i < mod->ins; i++) {
+		smplarr[i] = hio_read32b(f);
+		if (hio_eof(f)) {
+			D_(D_CRIT "read error at smplarr pos %d", i);
+			goto err_cleanup;
+		}
+	}
+
 	/*
 	 * Obtain number of samples from each instrument
 	 */
 	mod->smp = 0;
 	for (i = 0; i < mod->ins; i++) {
-		uint32 smpl_offset;
 		int16 type;
-		if (hio_seek(f, start + smplarr_offset + i * 4, SEEK_SET) != 0)
-		  return -1;
-		smpl_offset = hio_read32b(f);
-		if (smpl_offset == 0)
+		if (smplarr[i] == 0)
 			continue;
-		if (hio_seek(f, start + smpl_offset, SEEK_SET) != 0)
-		  return -1;
+		if (hio_seek(f, start + smplarr[i], SEEK_SET) != 0) {
+			D_(D_CRIT "seek error at instrument %d", i);
+			goto err_cleanup;
+		}
+
 		hio_read32b(f);				/* length */
 		type = hio_read16b(f);
-		if (type == -1) {			/* type is synth? */
+		if (type == -1 || type == -2) {		/* type is synth? */
 			int wforms;
 			hio_seek(f, 14, SEEK_CUR);
 			wforms = hio_read16b(f);
 
 			/* Sanity check */
-			if (wforms > 256)
-				return -1;
+			if (wforms > 256) {
+				D_(D_CRIT "invalid wform count at instrument %d", i);
+				goto err_cleanup;
+			}
 
 			mod->smp += wforms;
 		} else if (type >= 1 && type <= 6) {
-			mod->smp += num_oct[type - 1];
+			mod->smp += mmd_num_oct[type - 1];
 		} else {
 			mod->smp++;
 		}
@@ -229,8 +258,10 @@ static int mmd1_load(struct module_data *m, HIO_HANDLE *f, const int start)
 	expsmp_offset = 0;
 	iinfo_offset = 0;
 	if (expdata_offset) {
-		if (hio_seek(f, start + expdata_offset, SEEK_SET) != 0)
-		  return -1;
+		if (hio_seek(f, start + expdata_offset, SEEK_SET) != 0) {
+			D_(D_CRIT "seek error at expdata");
+			goto err_cleanup;
+		}
 		hio_read32b(f);
 		expsmp_offset = hio_read32b(f);
 		D_(D_INFO "expsmp_offset = 0x%08x", expsmp_offset);
@@ -246,9 +277,11 @@ static int mmd1_load(struct module_data *m, HIO_HANDLE *f, const int start)
 		/* Sanity check */
 		if (expsmp_offset < 0 ||
 		    annotxt_offset < 0 ||
-                    expdata.annolen > 0x10000 ||
+		    expdata.annolen > 0x10000 ||
 		    iinfo_offset < 0) {
-			return -1;
+			D_(D_CRIT "invalid expdata (annotxt=0x%08x annolen=0x%08x)",
+			   annotxt_offset, expdata.annolen);
+			goto err_cleanup;
 		}
 
 		hio_read32b(f);
@@ -270,7 +303,7 @@ static int mmd1_load(struct module_data *m, HIO_HANDLE *f, const int start)
 		/* Read annotation */
 		if (annotxt_offset != 0 && expdata.annolen != 0) {
 			D_(D_INFO "annotxt_offset = 0x%08x", annotxt_offset);
-			m->comment = malloc(expdata.annolen + 1);
+			m->comment = (char *) malloc(expdata.annolen + 1);
 			if (m->comment != NULL) {
 				hio_seek(f, start + annotxt_offset, SEEK_SET);
 				hio_read(m->comment, 1, expdata.annolen, f);
@@ -279,45 +312,75 @@ static int mmd1_load(struct module_data *m, HIO_HANDLE *f, const int start)
 		}
 	}
 
+	/*
+	 * Read blockarr.
+	 */
+	D_(D_WARN "read blockarr");
+	blockarr = (uint32 *) malloc(mod->pat * sizeof(uint32));
+	if (blockarr == NULL) {
+		goto err_cleanup;
+	}
+	if (hio_seek(f, start + blockarr_offset, SEEK_SET) != 0) {
+		D_(D_CRIT "seek error at blockarr");
+		goto err_cleanup;
+	}
+	for (i = 0; i < mod->pat; i++) {
+		blockarr[i] = hio_read32b(f);
+		if (hio_error(f)) {
+			D_(D_CRIT "read error at blockarr pos %d", i);
+			goto err_cleanup;
+		}
+	}
+
 	/*
 	 * Quickly scan patterns to check the number of channels
 	 */
 	D_(D_WARN "find number of channels");
 
+	max_lines = 1;
 	for (i = 0; i < mod->pat; i++) {
-		int block_offset;
-
-		if (hio_seek(f, start + blockarr_offset + i * 4, SEEK_SET) != 0)
-			return -1;
-		block_offset = hio_read32b(f);
-		D_(D_INFO "block %d block_offset = 0x%08x", i, block_offset);
-		if (block_offset == 0)
+		D_(D_INFO "block %d block_offset = 0x%08x", i, blockarr[i]);
+		if (blockarr[i] == 0)
 			continue;
-		if (hio_seek(f, start + block_offset, SEEK_SET) != 0)
-			return -1;
+
+		if (hio_seek(f, start + blockarr[i], SEEK_SET) != 0) {
+			D_(D_CRIT "seek error at block %d", i);
+			goto err_cleanup;
+		}
 
 		if (ver > 0) {
 			block.numtracks = hio_read16b(f);
-			/* block.lines = */ hio_read16b(f);
+			block.lines = hio_read16b(f);
 		} else {
 			block.numtracks = hio_read8(f);
-			/* block.lines = */ hio_read8(f);
+			block.lines = hio_read8(f);
+		}
+
+		/* Sanity check--Amiga OctaMED files have an upper bound of 3200 lines per block. */
+		if (block.lines + 1 > 3200) {
+			D_(D_CRIT "invalid line count %d in block %d", block.lines + 1, i);
+			goto err_cleanup;
 		}
 
 		if (block.numtracks > mod->chn) {
 			mod->chn = block.numtracks;
 		}
+		if (block.lines + 1 > max_lines) {
+			max_lines = block.lines + 1;
+		}
 	}
 
 	/* Sanity check */
 	/* MMD0/MMD1 can't have more than 16 channels... */
 	if (mod->chn > MIN(16, XMP_MAX_CHANNELS)) {
-		return -1;
+		D_(D_CRIT "invalid channel count %d", mod->chn);
+		goto err_cleanup;
 	}
 
 	mod->trk = mod->pat * mod->chn;
 
-	libxmp_set_type(m, ver == 0 ? mod->chn > 4 ? "OctaMED 2.00 MMD0" :
+	libxmp_set_type(m, ver == 0 ? mmdc ? "MED Packer MMDC" :
+				mod->chn > 4 ? "OctaMED 2.00 MMD0" :
 				"MED 2.10 MMD0" : "OctaMED 4.00 MMD1");
 
 	MODULE_INFO();
@@ -331,18 +394,23 @@ static int mmd1_load(struct module_data *m, HIO_HANDLE *f, const int start)
 	 */
 	D_(D_WARN "read patterns");
 	if (libxmp_init_pattern(mod) < 0)
-		return -1;
+		goto err_cleanup;
+
+	if ((patbuf = (uint8 *)malloc(mod->chn * max_lines * 4)) == NULL) {
+		goto err_cleanup;
+	}
 
 	for (i = 0; i < mod->pat; i++) {
-		int block_offset;
+		uint8 *pos;
+		size_t size;
 
-		if (hio_seek(f, start + blockarr_offset + i * 4, SEEK_SET) != 0)
-		  return -1;
-		block_offset = hio_read32b(f);
-		if (block_offset == 0)
+		if (blockarr[i] == 0)
 			continue;
-		if (hio_seek(f, start + block_offset, SEEK_SET) != 0)
-		  return -1;
+
+		if (hio_seek(f, start + blockarr[i], SEEK_SET) != 0) {
+			D_(D_CRIT "seek error at block %d", i);
+			goto err_cleanup;
+		}
 
 		if (ver > 0) {
 			block.numtracks = hio_read16b(f);
@@ -353,23 +421,50 @@ static int mmd1_load(struct module_data *m, HIO_HANDLE *f, const int start)
 			block.lines = hio_read8(f);
 		}
 
-		/* Sanity check--Amiga OctaMED files have an upper bound of 3200 lines per block. */
-		if (block.lines + 1 > 3200)
-			return -1;
+		size = block.numtracks * (block.lines + 1) * (ver ? 4 : 3);
+
+		if (mmdc) {
+			/* MMDC is just MMD0 with simple pattern packing. */
+			memset(patbuf, 0, size);
+			for (j = 0; j < size;) {
+				unsigned pack = hio_read8(f);
+				if (hio_error(f)) {
+					D_(D_CRIT "read error in block %d", i);
+					goto err_cleanup;
+				}
+
+				if (pack & 0x80) {
+					/* Run of 0 */
+					j += 256 - pack;
+					continue;
+				}
+				/* Uncompressed block */
+				pack++;
+				if (pack > size - j)
+					pack = size - j;
+
+				if (hio_read(patbuf + j, 1, pack, f) < pack) {
+					D_(D_CRIT "read error in block %d", i);
+					goto err_cleanup;
+				}
+				j += pack;
+			}
+		} else {
+			if (hio_read(patbuf, 1, size, f) < size) {
+				D_(D_CRIT "read error in block %d", i);
+				goto err_cleanup;
+			}
+		}
 
 		if (libxmp_alloc_pattern_tracks_long(mod, i, block.lines + 1) < 0)
-			return -1;
+			goto err_cleanup;
 
+		pos = patbuf;
 		if (ver > 0) {		/* MMD1 */
 			for (j = 0; j < mod->xxp[i]->rows; j++) {
 				for (k = 0; k < block.numtracks; k++) {
-					e[0] = hio_read8(f);
-					e[1] = hio_read8(f);
-					e[2] = hio_read8(f);
-					e[3] = hio_read8(f);
-
 					event = &EVENT(i, k, j);
-					event->note = e[0] & 0x7f;
+					event->note = pos[0] & 0x7f;
 					if (event->note)
 						event->note +=
 						    12 + song.playtransp;
@@ -377,182 +472,142 @@ static int mmd1_load(struct module_data *m, HIO_HANDLE *f, const int start)
 					if (event->note >= XMP_MAX_KEYS)
 						event->note = 0;
 
-					event->ins = e[1] & 0x3f;
+					event->ins = pos[1] & 0x3f;
 
 					/* Decay */
 					if (event->ins && !event->note) {
 						event->f2t = FX_MED_HOLD;
 					}
 
-					event->fxt = e[2];
-					event->fxp = e[3];
+					event->fxt = pos[2];
+					event->fxp = pos[3];
 					mmd_xlat_fx(event, bpm_on, bpmlen,
 							med_8ch, hexvol);
+					pos += 4;
 				}
 			}
 		} else {		/* MMD0 */
 			for (j = 0; j < mod->xxp[i]->rows; j++) {
 				for (k = 0; k < block.numtracks; k++) {
-					e[0] = hio_read8(f);
-					e[1] = hio_read8(f);
-					e[2] = hio_read8(f);
-
 					event = &EVENT(i, k, j);
-					event->note = e[0] & 0x3f;
+					event->note = pos[0] & 0x3f;
 					if (event->note)
-						event->note += 12;
+						event->note += 12 + song.playtransp;
 
 					if (event->note >= XMP_MAX_KEYS)
 						event->note = 0;
 
 					event->ins =
-					    (e[1] >> 4) | ((e[0] & 0x80) >> 3)
-					    | ((e[0] & 0x40) >> 1);
+					    (pos[1] >> 4) | ((pos[0] & 0x80) >> 3)
+					    | ((pos[0] & 0x40) >> 1);
 
 					/* Decay */
 					if (event->ins && !event->note) {
 						event->f2t = FX_MED_HOLD;
 					}
 
-					event->fxt = e[1] & 0x0f;
-					event->fxp = e[2];
+					event->fxt = pos[1] & 0x0f;
+					event->fxp = pos[2];
 					mmd_xlat_fx(event, bpm_on, bpmlen,
 							med_8ch, hexvol);
+					pos += 3;
 				}
 			}
 		}
 	}
+	free(patbuf);
+	patbuf = NULL;
 
 	if (libxmp_med_new_module_extras(m))
-		return -1;
+		goto err_cleanup;
 
 	/*
 	 * Read and convert instruments and samples
 	 */
 	D_(D_WARN "read instruments");
 	if (libxmp_init_instrument(m) < 0)
-		return -1;
+		goto err_cleanup;
 
 	D_(D_INFO "Instruments: %d", mod->ins);
 
-	for (smp_idx = i = 0; i < mod->ins; i++) {
-		int smpl_offset;
+	/* Instrument extras */
+	exp_smp = (struct InstrExt *) calloc(mod->ins, sizeof(struct InstrExt));
+	if (exp_smp == NULL) {
+		goto err_cleanup;
+	}
 
-		if (hio_seek(f, start + smplarr_offset + i * 4, SEEK_SET) != 0)
-			return -1;
-		smpl_offset = hio_read32b(f);
-
-		D_(D_INFO "sample %d smpl_offset = 0x%08x", i, smpl_offset);
-
-		if (smpl_offset == 0) {
-			continue;
-		}
-		if (hio_seek(f, start + smpl_offset, SEEK_SET) < 0) {
-			return -1;
-		}
-		instr.length = hio_read32b(f);
-		instr.type = hio_read16b(f);
-
-		if ((pos = hio_tell(f)) < 0) {
-			return -1;
+	if (expsmp_offset) {
+		if (hio_seek(f, start + expsmp_offset, SEEK_SET) != 0) {
+			D_(D_CRIT "seek error at expsmp");
+			goto err_cleanup;
 		}
 
-		if (expdata_offset && i < expdata.i_ext_entries) {
-			struct xmp_instrument *xxi = &mod->xxi[i];
-			int offset = iinfo_offset + i * expdata.i_ext_entrsz;
-			D_(D_INFO "sample %d iinfo_offset = 0x%08x", i, offset);
+		for (i = 0; i < mod->ins && i < expdata.s_ext_entries; i++) {
+			int skip = expdata.s_ext_entrsz - 4;
 
-			if (offset < 0 || hio_seek(f, start + offset, SEEK_SET) < 0) {
-				return -1;
+			D_(D_INFO "sample %d expsmp_offset = 0x%08lx", i, hio_tell(f));
+
+			exp_smp[i].hold = hio_read8(f);
+			exp_smp[i].decay = hio_read8(f);
+			exp_smp[i].suppress_midi_off = hio_read8(f);
+			exp_smp[i].finetune = hio_read8(f);
+
+			if (hio_error(f)) {
+				D_(D_CRIT "read error at expsmp");
+				goto err_cleanup;
 			}
+
+			if (skip && hio_seek(f, skip, SEEK_CUR) != 0) {
+				D_(D_CRIT "seek error at expsmp");
+				goto err_cleanup;
+			}
+		}
+	}
+
+	/* Instrument names */
+	if (iinfo_offset) {
+		uint8 name[40];
+
+		if (hio_seek(f, start + iinfo_offset, SEEK_SET) != 0) {
+			D_(D_CRIT "seek error at iinfo");
+			goto err_cleanup;
+		}
+
+		for (i = 0; i < mod->ins && i < expdata.i_ext_entries; i++) {
+			int skip = expdata.i_ext_entrsz - 40;
+
+			D_(D_INFO "sample %d iinfo_offset = 0x%08lx", i, hio_tell(f));
+
 			if (hio_read(name, 40, 1, f) < 1) {
 				D_(D_CRIT "read error at iinfo %d", i);
-				return -1;
+				goto err_cleanup;
 			}
-			strncpy(xxi->name, name, 32);
-			xxi->name[31] = '\0';
+			libxmp_instrument_name(mod, i, name, 40);
+
+			if (skip && hio_seek(f, skip, SEEK_CUR) != 0) {
+				D_(D_CRIT "seek error at iinfo %d", i);
+				goto err_cleanup;
+			}
+		}
+	}
+
+	/* Sample data */
+	for (smp_idx = i = 0; i < mod->ins; i++) {
+		D_(D_INFO "sample %d smpl_offset = 0x%08x", i, smplarr[i]);
+		if (smplarr[i] == 0) {
+			continue;
 		}
 
-		D_(D_INFO "[%2x] %-40.40s %d", i, mod->xxi[i].name, instr.type);
-
-		exp_smp.finetune = 0;
-		if (expdata_offset && i < expdata.s_ext_entries) {
-			int offset = expsmp_offset + i * expdata.s_ext_entrsz;
-			D_(D_INFO "sample %d expsmp_offset = 0x%08x", i, offset);
-
-			if (offset < 0 || hio_seek(f, start + offset, SEEK_SET) < 0) {
-				return -1;
-			}
-			exp_smp.hold = hio_read8(f);
-			exp_smp.decay = hio_read8(f);
-			exp_smp.suppress_midi_off = hio_read8(f);
-			exp_smp.finetune = hio_read8(f);
+		if (hio_seek(f, start + smplarr[i], SEEK_SET) < 0) {
+			D_(D_CRIT "seek error at instrument %d", i);
+			goto err_cleanup;
 		}
 
-		hio_seek(f, pos, SEEK_SET);
+		smp_idx = mmd_load_instrument(f, m, i, smp_idx, &expdata,
+			&exp_smp[i], &song.sample[i], ver);
 
-		if (instr.type == -2) {			/* Hybrid */
-			int ret = mmd_load_hybrid_instrument(f, m, i, smp_idx,
-				&synth, &exp_smp, &song.sample[i]);
-
-			smp_idx++;
-
-			if (ret < 0)
-				return -1;
-
-			if (mmd_alloc_tables(m, i, &synth) != 0)
-				return -1;
-
-			continue;
-		} else if (instr.type == -1) {			/* Synthetic */
-			int ret = mmd_load_synth_instrument(f, m, i, smp_idx,
-				&synth, &exp_smp, &song.sample[i]);
-
-			if (ret > 0)
-				continue;
-
-			if (ret < 0)
-				return -1;
-
-			smp_idx += synth.wforms;
-
-			if (mmd_alloc_tables(m, i, &synth) != 0)
-				return -1;
-
-			continue;
-		} else if (instr.type >= 1 && instr.type <= 6) { /* IFFOCT */
-			int ret;
-			const int oct = num_oct[instr.type - 1];
-
-			hio_seek(f, start + smpl_offset + 6, SEEK_SET);
-
-			ret = mmd_load_iffoct_instrument(f, m, i, smp_idx,
-				&instr, oct, &exp_smp, &song.sample[i]);
-
-			if (ret < 0)
-				return -1;
-
-			smp_idx += oct;
-
-			continue;
-		} else if (instr.type == 0) {			/* Sample */
-			int ret;
-
-			hio_seek(f, start + smpl_offset + 6, SEEK_SET);
-
-			ret = mmd_load_sampled_instrument(f, m, i, smp_idx,
-				&instr, &expdata, &exp_smp, &song.sample[i],
-				ver);
-
-			if (ret < 0)
-				return -1;
-
-			smp_idx++;
-
-			continue;
-		} else {
-			/* Invalid instrument type */
-			return -1;
+		if (smp_idx < 0) {
+			goto err_cleanup;
 		}
 	}
 
@@ -562,6 +617,13 @@ static int mmd1_load(struct module_data *m, HIO_HANDLE *f, const int start)
 	}
 
 	m->read_event_type = READ_EVENT_MED;
+	retval = 0;
 
-	return 0;
+    err_cleanup:
+	free(exp_smp);
+	free(blockarr);
+	free(smplarr);
+	free(patbuf);
+
+	return retval;
 }
