@@ -1,5 +1,5 @@
 /* Extended Module Player
- * Copyright (C) 1996-2021 Claudio Matsuoka and Hipolito Carraro Jr
+ * Copyright (C) 1996-2022 Claudio Matsuoka and Hipolito Carraro Jr
  *
  * Permission is hereby granted, free of charge, to any person obtaining a
  * copy of this software and associated documentation files (the "Software"),
@@ -22,7 +22,7 @@
 
 #include "med.h"
 #include "loader.h"
-#include "med_extras.h"
+#include "../med_extras.h"
 
 static int mmd3_test (HIO_HANDLE *, char *, const int);
 static int mmd3_load (struct module_data *, HIO_HANDLE *, const int);
@@ -60,8 +60,6 @@ static int mmd3_test(HIO_HANDLE *f, char *t, const int start)
 	return 0;
 }
 
-/* Number of octaves in IFFOCT samples */
-static const int num_oct[6] = { 5, 3, 2, 4, 6, 7 };
 
 static int mmd3_load(struct module_data *m, HIO_HANDLE *f, const int start)
 {
@@ -70,14 +68,14 @@ static int mmd3_load(struct module_data *m, HIO_HANDLE *f, const int start)
 	struct MMD0 header;
 	struct MMD2song song;
 	struct MMD1Block block;
-	struct InstrHdr instr;
-	struct SynthInstr synth;
-	struct InstrExt exp_smp;
+	struct InstrExt *exp_smp = NULL;
 	struct MMD0exp expdata;
 	struct xmp_event *event;
+	uint32 *blockarr = NULL;
+	uint32 *smplarr = NULL;
+	uint8 *patbuf = NULL;
 	int ver = 0;
 	int smp_idx = 0;
-	uint8 e[4];
 	int song_offset;
 	int seqtable_offset;
 	int trackvols_offset;
@@ -90,9 +88,9 @@ static int mmd3_load(struct module_data *m, HIO_HANDLE *f, const int start)
 	int iinfo_offset;
 	int mmdinfo_offset;
 	int playseq_offset;
-	int pos;
 	int bpm_on, bpmlen, med_8ch, hexvol;
-	char name[40];
+	int max_lines;
+	int retval = -1;
 
 	LOAD_INIT();
 
@@ -167,6 +165,7 @@ static int mmd3_load(struct module_data *m, HIO_HANDLE *f, const int start)
 
 	/* Sanity check */
 	if (song.numsamples > 63) {
+		D_(D_CRIT "invalid instrument count %d", song.numsamples);
 		return -1;
 	}
 
@@ -183,6 +182,7 @@ static int mmd3_load(struct module_data *m, HIO_HANDLE *f, const int start)
 
 	/* Sanity check */
 	if (mod->len > 255) {
+		D_(D_CRIT "unsupported song length %d", mod->len);
 		return -1;
 	}
 
@@ -210,26 +210,49 @@ static int mmd3_load(struct module_data *m, HIO_HANDLE *f, const int start)
 	mod->chn = 0;
 	mod->name[0] = 0;
 
+	/*
+	 * Read smplarr
+	 */
+	D_(D_WARN "read smplarr");
+	smplarr = (uint32 *) malloc(mod->ins * sizeof(uint32));
+	if (smplarr == NULL) {
+		return -1;
+	}
+	if (hio_seek(f, start + smplarr_offset, SEEK_SET) != 0) {
+		D_(D_CRIT "seek error at smplarr");
+		goto err_cleanup;
+	}
+	for (i = 0; i < mod->ins; i++) {
+		smplarr[i] = hio_read32b(f);
+		if (hio_eof(f)) {
+			D_(D_CRIT "read error at smplarr pos %d", i);
+			goto err_cleanup;
+		}
+	}
+
 	/*
 	 * Obtain number of samples from each instrument
 	 */
 	mod->smp = 0;
 	for (i = 0; i < mod->ins; i++) {
-		uint32 smpl_offset;
 		int16 type;
-		hio_seek(f, start + smplarr_offset + i * 4, SEEK_SET);
-		smpl_offset = hio_read32b(f);
-		if (smpl_offset == 0)
+		if (smplarr[i] == 0)
 			continue;
-		hio_seek(f, start + smpl_offset, SEEK_SET);
+		hio_seek(f, start + smplarr[i], SEEK_SET);
 		hio_read32b(f);				/* length */
 		type = hio_read16b(f);
-		if (type == -1) {			/* type is synth? */
+		if (type == -1 || type == -2) {		/* type is synth? */
 			hio_seek(f, 14, SEEK_CUR);
 			mod->smp += hio_read16b(f);	/* wforms */
+		} else if (type >= 1 && type <= 6) {	/* octave samples */
+			mod->smp += mmd_num_oct[type - 1];
 		} else {
 			mod->smp++;
 		}
+		if (hio_error(f)) {
+			D_(D_CRIT "read error at sample %d", i);
+			goto err_cleanup;
+		}
 	}
 
 	/*
@@ -244,7 +267,10 @@ static int mmd3_load(struct module_data *m, HIO_HANDLE *f, const int start)
 	iinfo_offset = 0;
 
 	if (expdata_offset) {
-		hio_seek(f, start + expdata_offset, SEEK_SET);
+		if (hio_seek(f, start + expdata_offset, SEEK_SET) != 0) {
+			D_(D_CRIT "seek error at expdata");
+			goto err_cleanup;
+		}
 		hio_read32b(f);
 		expsmp_offset = hio_read32b(f);
 		D_(D_INFO "expsmp_offset = 0x%08x", expsmp_offset);
@@ -259,7 +285,8 @@ static int mmd3_load(struct module_data *m, HIO_HANDLE *f, const int start)
 
 		/* Sanity check */
 		if (expsmp_offset < 0 || iinfo_offset < 0) {
-			return -1;
+			D_(D_CRIT "invalid expdata");
+			goto err_cleanup;
 		}
 
 		hio_read32b(f);
@@ -273,7 +300,8 @@ static int mmd3_load(struct module_data *m, HIO_HANDLE *f, const int start)
 		mmdinfo_offset = hio_read32b(f);
 
 		if (hio_error(f)) {
-			return -1;
+			D_(D_CRIT "read error in expdata");
+			goto err_cleanup;
 		}
 
 		hio_seek(f, start + songname_offset, SEEK_SET);
@@ -291,38 +319,67 @@ static int mmd3_load(struct module_data *m, HIO_HANDLE *f, const int start)
 		}
 	}
 
+	/*
+	 * Read blockarr.
+	 */
+	D_(D_WARN "read blockarr");
+	blockarr = (uint32 *) malloc(mod->pat * sizeof(uint32));
+	if (blockarr == NULL) {
+		goto err_cleanup;
+	}
+	if (hio_seek(f, start + blockarr_offset, SEEK_SET) != 0) {
+		D_(D_CRIT "seek error at blockarr");
+		goto err_cleanup;
+	}
+	for (i = 0; i < mod->pat; i++) {
+		blockarr[i] = hio_read32b(f);
+		if (hio_error(f)) {
+			D_(D_CRIT "read error at blockarr pos %d", i);
+			goto err_cleanup;
+		}
+	}
+
 	/*
 	 * Quickly scan patterns to check the number of channels
 	 */
 	D_(D_WARN "find number of channels");
 
+	max_lines = 1;
 	for (i = 0; i < mod->pat; i++) {
-		int block_offset;
-
-		hio_seek(f, start + blockarr_offset + i * 4, SEEK_SET);
-		block_offset = hio_read32b(f);
-		D_(D_INFO "block %d block_offset = 0x%08x", i, block_offset);
-		if (hio_error(f)) {
-			return -1;
-		}
-		if (block_offset == 0)
+		D_(D_INFO "block %d block_offset = 0x%08x", i, blockarr[i]);
+		if (blockarr[i] == 0)
 			continue;
-		hio_seek(f, start + block_offset, SEEK_SET);
+
+		hio_seek(f, start + blockarr[i], SEEK_SET);
 
 		block.numtracks = hio_read16b(f);
-		/* block.lines = */ hio_read16b(f);
+		block.lines = hio_read16b(f);
 		if (hio_error(f)) {
-			return -1;
+			D_(D_CRIT "read error at block %d", i);
+			goto err_cleanup;
+		}
+
+		/* Sanity check--Amiga OctaMED files have an upper bound of 3200 lines per block,
+		 * but MED Soundstudio for Windows allows up to 9999 lines.
+		  */
+		if (block.lines + 1 > 9999) {
+			D_(D_CRIT "invalid line count %d in block %d", block.lines + 1, i);
+			goto err_cleanup;
 		}
 
 		if (block.numtracks > mod->chn) {
 			mod->chn = block.numtracks;
 		}
+		if (block.lines + 1 > max_lines) {
+			max_lines = block.lines + 1;
+		}
 	}
 
 	/* Sanity check */
-	if (mod->chn <= 0 || mod->chn > XMP_MAX_CHANNELS)
-		return -1;
+	if (mod->chn <= 0 || mod->chn > XMP_MAX_CHANNELS) {
+		D_(D_CRIT "invalid channel count %d", mod->chn);
+		goto err_cleanup;
+	}
 
 	mod->trk = mod->pat * mod->chn;
 
@@ -343,67 +400,66 @@ static int mmd3_load(struct module_data *m, HIO_HANDLE *f, const int start)
 	D_(D_WARN "read patterns");
 
 	if (libxmp_init_pattern(mod) < 0)
-		return -1;
+		goto err_cleanup;
+
+	if ((patbuf = (uint8 *)malloc(mod->chn * max_lines * 4)) == NULL) {
+		goto err_cleanup;
+	}
 
 	for (i = 0; i < mod->pat; i++) {
-		int block_offset;
+		uint8 *pos;
+		size_t size;
 
-		hio_seek(f, start + blockarr_offset + i * 4, SEEK_SET);
-		block_offset = hio_read32b(f);
-		if (block_offset == 0)
+		if (blockarr[i] == 0)
 			continue;
-		hio_seek(f, start + block_offset, SEEK_SET);
+
+		hio_seek(f, start + blockarr[i], SEEK_SET);
 
 		block.numtracks = hio_read16b(f);
 		block.lines = hio_read16b(f);
-		hio_read32b(f);
+		hio_read32b(f); /* FIXME: should try to load extra command pages when they exist. */
 
-		/* Sanity check--Amiga OctaMED files have an upper bound of 3200 lines per block,
-		 * but MED Soundstudio for Windows allows up to 9999 lines.
-		  */
-		if (block.lines + 1 > 9999)
-			return -1;
+		size = block.numtracks * (block.lines + 1) * 4;
+		if (hio_read(patbuf, 1, size, f) < size) {
+			D_(D_CRIT "read error in block %d", i);
+			goto err_cleanup;
+		}
 
 		if (libxmp_alloc_pattern_tracks_long(mod, i, block.lines + 1) < 0)
-			return -1;
+			goto err_cleanup;
 
+		pos = patbuf;
 		for (j = 0; j < mod->xxp[i]->rows; j++) {
 			for (k = 0; k < block.numtracks; k++) {
-				e[0] = hio_read8(f);
-				e[1] = hio_read8(f);
-				e[2] = hio_read8(f);
-				e[3] = hio_read8(f);
-
 				event = &EVENT(i, k, j);
-				event->note = e[0] & 0x7f;
+				event->note = pos[0] & 0x7f;
 				if (event->note) {
-					event->note += song.playtransp;
-					if (ver == 2)
-						event->note += 12;
-					else
-						event->note -= 12;
-				};
+					event->note += 12 + song.playtransp;
+				}
 
 				if (event->note >= XMP_MAX_KEYS)
 					event->note = 0;
 
-				event->ins = e[1] & 0x3f;
+				event->ins = pos[1] & 0x3f;
 
 				/* Decay */
 				if (event->ins && !event->note) {
 					event->f2t = FX_MED_HOLD;
 				}
 
-				event->fxt = e[2];
-				event->fxp = e[3];
+				event->fxt = pos[2];
+				event->fxp = pos[3];
 				mmd_xlat_fx(event, bpm_on, bpmlen,
 						med_8ch, hexvol);
+				pos += 4;
 			}
 		}
 	}
+	free(patbuf);
+	patbuf = NULL;
 
 	if (libxmp_med_new_module_extras(m) != 0)
-		return -1;
+		goto err_cleanup;
 
 	/*
 	 * Read and convert instruments and samples
@@ -411,136 +467,104 @@ static int mmd3_load(struct module_data *m, HIO_HANDLE *f, const int start)
 	D_(D_WARN "read instruments");
 
 	if (libxmp_init_instrument(m) < 0)
-		return -1;
+		goto err_cleanup;
 
 	D_(D_INFO "Instruments: %d", mod->ins);
 
-	for (smp_idx = i = 0; i < mod->ins; i++) {
-		int smpl_offset;
+	/* Instrument extras */
+	exp_smp = (struct InstrExt *) calloc(mod->ins, sizeof(struct InstrExt));
+	if (exp_smp == NULL) {
+		goto err_cleanup;
+	}
 
-		hio_seek(f, start + smplarr_offset + i * 4, SEEK_SET);
-		smpl_offset = hio_read32b(f);
-
-		D_(D_INFO "sample %d smpl_offset = 0x%08x", i, smpl_offset);
-
-		if (smpl_offset == 0) {
-			continue;
+	if (expsmp_offset) {
+		if (hio_seek(f, start + expsmp_offset, SEEK_SET) != 0) {
+			D_(D_CRIT "seek error at expsmp");
+			goto err_cleanup;
 		}
 
-		hio_seek(f, start + smpl_offset, SEEK_SET);
-		instr.length = hio_read32b(f);
-		instr.type = hio_read16b(f);
+		for (i = 0; i < mod->ins && i < expdata.s_ext_entries; i++) {
+			int skip = expdata.s_ext_entrsz - 4;
 
-		if ((pos = hio_tell(f)) < 0) {
-			return -1;
-		}
+			D_(D_INFO "sample %d expsmp_offset = 0x%08lx", i, hio_tell(f));
 
-		if (expdata_offset && i < expdata.i_ext_entries) {
-			struct xmp_instrument *xxi = &mod->xxi[i];
-			int offset = iinfo_offset + i * expdata.i_ext_entrsz;
-			D_(D_INFO "sample %d iinfo_offset = 0x%08x", i, offset);
+			exp_smp[i].hold = hio_read8(f);
+			exp_smp[i].decay = hio_read8(f);
+			exp_smp[i].suppress_midi_off = hio_read8(f);
+			exp_smp[i].finetune = hio_read8(f);
 
-			if (offset < 0 || hio_seek(f, start + offset, SEEK_SET) < 0) {
-				return -1;
+			if (expdata.s_ext_entrsz >= 8) {	/* Octamed V5 */
+				exp_smp[i].default_pitch = hio_read8(f);
+				exp_smp[i].instr_flags = hio_read8(f);
+				hio_read16b(f);
+				skip -= 4;
 			}
+			if (expdata.s_ext_entrsz >= 10) {	/* OctaMED V5.02 */
+				hio_read16b(f);
+				skip -= 2;
+			}
+			if (expdata.s_ext_entrsz >= 18) {	/* OctaMED V7 */
+				exp_smp[i].long_repeat = hio_read32b(f);
+				exp_smp[i].long_replen = hio_read32b(f);
+				skip -= 8;
+			}
+
+			if (hio_error(f)) {
+				D_(D_CRIT "read error at expsmp");
+				goto err_cleanup;
+			}
+
+			if (skip && hio_seek(f, skip, SEEK_CUR) != 0) {
+				D_(D_CRIT "seek error at expsmp");
+				goto err_cleanup;
+			}
+		}
+	}
+
+	/* Instrument names */
+	if (iinfo_offset) {
+		uint8 name[40];
+
+		if (hio_seek(f, start + iinfo_offset, SEEK_SET) != 0) {
+			D_(D_CRIT "seek error at iinfo");
+			goto err_cleanup;
+		}
+
+		for (i = 0; i < mod->ins && i < expdata.i_ext_entries; i++) {
+			int skip = expdata.i_ext_entrsz - 40;
+
+			D_(D_INFO "sample %d iinfo_offset = 0x%08lx", i, hio_tell(f));
+
 			if (hio_read(name, 40, 1, f) < 1) {
 				D_(D_CRIT "read error at iinfo %d", i);
-				return -1;
+				goto err_cleanup;
 			}
-			strncpy(xxi->name, name, 32);
-			xxi->name[31] = '\0';
-			D_(D_INFO "[%2x] %-40.40s %d", i, mod->xxi[i].name, instr.type);
-		}
+			libxmp_instrument_name(mod, i, name, 40);
 
-		memset(&exp_smp, 0, sizeof(struct InstrExt));
-
-		if (expdata_offset && i < expdata.s_ext_entries) {
-			int offset = expsmp_offset + i * expdata.s_ext_entrsz;
-			D_(D_INFO "sample %d expsmp_offset = 0x%08x", i, offset);
-
-			if (offset < 0 || hio_seek(f, start + offset, SEEK_SET) < 0) {
-				return -1;
-			}
-			exp_smp.hold = hio_read8(f);
-			exp_smp.decay = hio_read8(f);
-			exp_smp.suppress_midi_off = hio_read8(f);
-			exp_smp.finetune = hio_read8(f);
-
-			if (expdata.s_ext_entrsz > 4) {	/* Octamed V5 */
-				exp_smp.default_pitch = hio_read8(f);
-				exp_smp.instr_flags = hio_read8(f);
+			if (skip && hio_seek(f, skip, SEEK_CUR) != 0) {
+				D_(D_CRIT "seek error at iinfo %d", i);
+				goto err_cleanup;
 			}
 		}
+	}
 
-		hio_seek(f, pos, SEEK_SET);
-
-		if (instr.type == -2) {			/* Hybrid */
-			int ret = mmd_load_hybrid_instrument(f, m, i,
-				smp_idx, &synth, &exp_smp, &song.sample[i]);
-
-			if (ret < 0)
-				return -1;
-
-			smp_idx++;
-
-			if (mmd_alloc_tables(m, i, &synth) != 0)
-				return -1;
-
+	/* Sample data */
+	for (smp_idx = i = 0; i < mod->ins; i++) {
+		D_(D_INFO "sample %d smpl_offset = 0x%08x", i, smplarr[i]);
+		if (smplarr[i] == 0) {
 			continue;
-		} else if (instr.type == -1) {		/* Synthetic */
-			int ret = mmd_load_synth_instrument(f, m, i, smp_idx,
-				&synth, &exp_smp, &song.sample[i]);
+		}
 
-			if (ret > 0)
-				continue;
+		if (hio_seek(f, start + smplarr[i], SEEK_SET) < 0) {
+			D_(D_CRIT "seek error at instrument %d", i);
+			goto err_cleanup;
+		}
 
-			if (ret < 0)
-				return -1;
+		smp_idx = mmd_load_instrument(f, m, i, smp_idx, &expdata,
+			&exp_smp[i], &song.sample[i], ver);
 
-			smp_idx += synth.wforms;
-
-			if (mmd_alloc_tables(m, i, &synth) != 0)
-				return -1;
-
-			continue;
-		} else if (instr.type >= 1 && instr.type <= 6) { /* IFFOCT */
-			int ret;
-			const int oct = num_oct[instr.type - 1];
-
-			hio_seek(f, start + smpl_offset + 6, SEEK_SET);
-
-			ret = mmd_load_iffoct_instrument(f, m, i, smp_idx,
-				&instr, oct, &exp_smp, &song.sample[i]);
-
-			if (ret < 0)
-				return -1;
-
-			smp_idx += oct;
-
-			continue;
-		} else if ((instr.type & STEREO) != 0) {
-			D_(D_WARN "stereo sample unsupported");
-			mod->xxi[i].nsm = 0;
-			continue;
-		} else if ((instr.type & S_16) == 0) {	/* Sample */
-			int ret;
-
-			hio_seek(f, start + smpl_offset + 6, SEEK_SET);
-
-			ret = mmd_load_sampled_instrument(f, m, i, smp_idx,
-				&instr, &expdata, &exp_smp, &song.sample[i],
-				ver);
-
-			if (ret < 0)
-				return -1;
-
-			smp_idx++;
-
-			continue;
-		} else {
-			/* Invalid instrument type */
-			D_(D_CRIT "invalid instrument type: %d", instr.type);
-			return -1;
+		if (smp_idx < 0) {
+			goto err_cleanup;
 		}
 	}
 
@@ -560,6 +584,13 @@ static int mmd3_load(struct module_data *m, HIO_HANDLE *f, const int start)
 	}
 
 	m->read_event_type = READ_EVENT_MED;
+	retval = 0;
 
-	return 0;
+    err_cleanup:
+	free(exp_smp);
+	free(blockarr);
+	free(smplarr);
+	free(patbuf);
+
+	return retval;
 }
