@@ -1,9 +1,31 @@
+/* ProWizard
+ * Copyright (C) 1997 Asle / ReDoX
+ * Modified in 2010,2014 by Claudio Matsuoka
+ * Modified in 2020 by Alice Rowan
+ *
+ * Permission is hereby granted, free of charge, to any person obtaining a
+ * copy of this software and associated documentation files (the "Software"),
+ * to deal in the Software without restriction, including without limitation
+ * the rights to use, copy, modify, merge, publish, distribute, sublicense,
+ * and/or sell copies of the Software, and to permit persons to whom the
+ * Software is furnished to do so, subject to the following conditions:
+ *
+ * The above copyright notice and this permission notice shall be included in
+ * all copies or substantial portions of the Software.
+ *
+ * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
+ * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
+ * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
+ * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
+ * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
+ * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
+ * THE SOFTWARE.
+ */
+
 /*
- *  NoiseRunner.c	Copyright (C) 1997 Asle / ReDoX
+ *  NoiseRunner.c
  *
  *  Converts NoiseRunner packed MODs back to Protracker
- *
- *  Modified in 2010,2014 by Claudio Matsuoka
  */
 
 #include "prowiz.h"
@@ -137,8 +159,15 @@ static int test_nru(const uint8 *data, char *t, int s)
 
 	/* test #3 volumes */
 	for (i = 0; i < 31; i++) {
+		int addr, start;
 		if (data[1 + i * 16] > 0x40)
 			return -1;
+
+		/* Also check sample offsets. */
+		addr = readmem32b(data + 2 + i * 16);
+		start = readmem32b(data + 8 + i * 16);
+		if (addr < 0 || start < 0 || start < addr)
+			return -1;
 	}
 
 	/* test #4  pattern list size */
