@@ -1,5 +1,5 @@
 /* Extended Module Player
- * Copyright (C) 1996-2021 Claudio Matsuoka and Hipolito Carraro Jr
+ * Copyright (C) 1996-2022 Claudio Matsuoka and Hipolito Carraro Jr
  *
  * Permission is hereby granted, free of charge, to any person obtaining a
  * copy of this software and associated documentation files (the "Software"),
@@ -20,11 +20,10 @@
  * THE SOFTWARE.
  */
 
-#include "common.h"
+#include "../common.h"
 #include "loader.h"
 
 #ifndef LIBXMP_CORE_PLAYER
-
 /*
  * From the Audio File Formats (version 2.5)
  * Submitted-by: Guido van Rossum <guido@cwi.nl>
@@ -55,7 +54,6 @@ static const int8 vdic_table[128] = {
 	/* 120 */	 95,  98, 103, 109, 114, 120, 126, 127
 };
 
-
 /* Convert 7 bit samples to 8 bit */
 static void convert_7bit_to_8bit(uint8 *p, int l)
 {
@@ -95,24 +93,23 @@ static void adpcm4_decoder(uint8 *inp, uint8 *outp, char *tab, int len)
 		*outp++ = delta;
 	}
 }
-
 #endif
 
 /* Convert differential to absolute sample data */
 static void convert_delta(uint8 *p, int l, int r)
 {
 	uint16 *w = (uint16 *)p;
-	uint16 abs = 0;
+	uint16 absval = 0;
 
 	if (r) {
 		for (; l--;) {
-			abs = *w + abs;
-			*w++ = abs;
+			absval = *w + absval;
+			*w++ = absval;
 		}
 	} else {
 		for (; l--;) {
-			abs = *p + abs;
-			*p++ = (uint8) abs;
+			absval = *p + absval;
+			*p++ = (uint8) absval;
 		}
 	}
 }
@@ -163,41 +160,10 @@ static void convert_stereo_to_mono(uint8 *p, int l, int r)
 }
 #endif
 
-static void unroll_loop(struct xmp_sample *xxs)
-{
-	int8 *s8;
-	int16 *s16;
-	int start, loop_size;
-	int i;
-
-	s16 = (int16 *)xxs->data;
-	s8 = (int8 *)xxs->data;
-
-	if (xxs->len > xxs->lpe) {
-		start = xxs->lpe;
-	} else {
-		start = xxs->len;
-	}
-
-	loop_size = xxs->lpe - xxs->lps;
-
-	if (xxs->flg & XMP_SAMPLE_16BIT) {
-		s16 += start;
-		for (i = 0; i < loop_size; i++) {
-			*(s16 + i) = *(s16 - i - 1);
-		}
-	} else {
-		s8 += start;
-		for (i = 0; i < loop_size; i++) {
-			*(s8 + i) = *(s8 - i - 1);
-		}
-	}
-}
-
 
 int libxmp_load_sample(struct module_data *m, HIO_HANDLE *f, int flags, struct xmp_sample *xxs, const void *buffer)
 {
-	int bytelen, extralen, unroll_extralen, i;
+	int bytelen, extralen, i;
 
 #ifndef LIBXMP_CORE_PLAYER
 	/* Adlib FM patches */
@@ -264,7 +230,6 @@ int libxmp_load_sample(struct module_data *m, HIO_HANDLE *f, int flags, struct x
 	 */
 	bytelen = xxs->len;
 	extralen = 4;
-	unroll_extralen = 0;
 
 	/* Disable birectional loop flag if sample is not looped
 	 */
@@ -272,25 +237,18 @@ int libxmp_load_sample(struct module_data *m, HIO_HANDLE *f, int flags, struct x
 		if (~xxs->flg & XMP_SAMPLE_LOOP)
 			xxs->flg &= ~XMP_SAMPLE_LOOP_BIDIR;
 	}
-	/* Unroll bidirectional loops
-	 */
-	if (xxs->flg & XMP_SAMPLE_LOOP_BIDIR) {
-		unroll_extralen = (xxs->lpe - xxs->lps) -
-				(xxs->len - xxs->lpe);
-
-		if (unroll_extralen < 0) {
-			unroll_extralen = 0;
-		}
+	if (xxs->flg & XMP_SAMPLE_SLOOP_BIDIR) {
+		if (~xxs->flg & XMP_SAMPLE_SLOOP)
+			xxs->flg &= ~XMP_SAMPLE_SLOOP_BIDIR;
 	}
 
 	if (xxs->flg & XMP_SAMPLE_16BIT) {
 		bytelen *= 2;
 		extralen *= 2;
-		unroll_extralen *= 2;
 	}
 
 	/* add guard bytes before the buffer for higher order interpolation */
-	xxs->data = malloc(bytelen + extralen + unroll_extralen + 4);
+	xxs->data = (unsigned char *) malloc(bytelen + extralen + 4);
 	if (xxs->data == NULL) {
 		goto err;
 	}
@@ -379,12 +337,6 @@ int libxmp_load_sample(struct module_data *m, HIO_HANDLE *f, int flags, struct x
 		xxs->flg |= XMP_SAMPLE_LOOP_FULL;
 	}
 
-	/* Unroll bidirectional loops */
-	if (xxs->flg & XMP_SAMPLE_LOOP_BIDIR) {
-		unroll_loop(xxs);
-		bytelen += unroll_extralen;
-	}
-
 	/* Add extra samples at end */
 	if (xxs->flg & XMP_SAMPLE_16BIT) {
 		for (i = 0; i < 8; i++) {
@@ -404,28 +356,6 @@ int libxmp_load_sample(struct module_data *m, HIO_HANDLE *f, int flags, struct x
 		xxs->data[-1] = xxs->data[0];
 	}
 
-	/* Fix sample at loop */
-	if (xxs->flg & XMP_SAMPLE_LOOP) {
-		int lpe = xxs->lpe;
-		int lps = xxs->lps;
-
-		if (xxs->flg & XMP_SAMPLE_LOOP_BIDIR) {
-			lpe += lpe - lps;
-		}
-
-		if (xxs->flg & XMP_SAMPLE_16BIT) {
-			lpe <<= 1;
-			lps <<= 1;
-			for (i = 0; i < 8; i++) {
-				xxs->data[lpe + i] = xxs->data[lps + i];
-			}
-		} else {
-			for (i = 0; i < 4; i++) {
-				xxs->data[lpe + i] = xxs->data[lps + i];
-			}
-		}
-	}
-
 	return 0;
 
 #ifndef LIBXMP_CORE_PLAYER
@@ -438,8 +368,8 @@ int libxmp_load_sample(struct module_data *m, HIO_HANDLE *f, int flags, struct x
 
 void libxmp_free_sample(struct xmp_sample *s)
 {
-    if (s->data) {
-	free(s->data - 4);
-	s->data = NULL;		/* prevent double free in PCM load error */
-    }
+	if (s->data) {
+		free(s->data - 4);
+		s->data = NULL;		/* prevent double free in PCM load error */
+	}
 }
