@@ -1,5 +1,5 @@
 /* Extended Module Player
- * Copyright (C) 1996-2021 Claudio Matsuoka and Hipolito Carraro Jr
+ * Copyright (C) 1996-2023 Claudio Matsuoka and Hipolito Carraro Jr
  *
  * Permission is hereby granted, free of charge, to any person obtaining a
  * copy of this software and associated documentation files (the "Software"),
@@ -42,6 +42,12 @@
 #include "effects.h"
 #include "mixer.h"
 
+#ifndef LIBXMP_CORE_PLAYER
+#include "far_extras.h"
+#endif
+
+#define VBLANK_TIME_THRESHOLD	480000 /* 8 minutes */
+
 #define S3M_END		0xff
 #define S3M_SKIP	0xfe
 
@@ -50,30 +56,35 @@ static int scan_module(struct context_data *ctx, int ep, int chain)
 {
     struct player_data *p = &ctx->p;
     struct module_data *m = &ctx->m;
-    struct xmp_module *mod = &m->mod;
+    const struct xmp_module *mod = &m->mod;
+    const struct xmp_track *tracks[XMP_MAX_CHANNELS];
+    const struct xmp_event *event;
     int parm, gvol_memory, f1, f2, p1, p2, ord, ord2;
     int row, last_row, break_row, row_count, row_count_total;
     int orders_since_last_valid, any_valid;
     int gvl, bpm, speed, base_time, chn;
     int frame_count;
     double time, start_time;
-    int loop_chn, loop_num, inside_loop;
+    int loop_chn, loop_num, inside_loop, line_jump;
     int pdelay = 0;
     int loop_count[XMP_MAX_CHANNELS];
     int loop_row[XMP_MAX_CHANNELS];
-    struct xmp_event* event;
     int i, pat;
     int has_marker;
     struct ord_data *info;
 #ifndef LIBXMP_CORE_PLAYER
     int st26_speed;
+    int far_tempo_coarse, far_tempo_fine, far_tempo_mode;
 #endif
+    /* was 255, but Global trash goes to 318.
+     * Higher limit for MEDs, defiance.crybaby.5 has blocks with 2048+ rows. */
+    const int row_limit = IS_PLAYER_MODE_MED() ? 3200 : 512;
 
     if (mod->len == 0)
 	return 0;
 
     for (i = 0; i < mod->len; i++) {
-	int pat = mod->xxo[i];
+	pat = mod->xxo[i];
 	memset(m->scan_cnt[i], 0, pat >= mod->pat ? 1 :
 			mod->xxp[pat]->rows ? mod->xxp[pat]->rows : 1);
     }
@@ -84,6 +95,7 @@ static int scan_module(struct context_data *ctx, int ep, int chain)
     }
     loop_num = 0;
     loop_chn = -1;
+    line_jump = 0;
 
     gvl = mod->gvl;
     bpm = mod->bpm;
@@ -92,6 +104,15 @@ static int scan_module(struct context_data *ctx, int ep, int chain)
     base_time = m->rrate;
 #ifndef LIBXMP_CORE_PLAYER
     st26_speed = 0;
+    far_tempo_coarse = 4;
+    far_tempo_fine = 0;
+    far_tempo_mode = 1;
+
+    if (HAS_FAR_MODULE_EXTRAS(ctx->m)) {
+	far_tempo_coarse = FAR_MODULE_EXTRAS(ctx->m)->coarse_tempo;
+	libxmp_far_translate_tempo(far_tempo_mode, 0, far_tempo_coarse,
+				   &far_tempo_fine, &speed, &bpm);
+    }
 #endif
 
     has_marker = HAS_QUIRK(QUIRK_MARKER);
@@ -196,6 +217,11 @@ static int scan_module(struct context_data *ctx, int ep, int chain)
 	    info->start_row = break_row;
 	}
 
+	/* Get tracks in advance to speed up the event parsing loop. */
+	for (chn = 0; chn < mod->chn; chn++) {
+		tracks[chn] = mod->xxt[TRACK_NUM(pat, chn)];
+	}
+
 	last_row = mod->xxp[pat]->rows;
 	for (row = break_row, break_row = 0; row < last_row; row++, row_count++, row_count_total++) {
 	    /* Prevent crashes caused by large softmixer frames */
@@ -216,12 +242,12 @@ static int scan_module(struct context_data *ctx, int ep, int chain)
 	     * (...) it dies at the end of position 2F
 	     */
 
-	    if (row_count_total > 512) { /* was 255, but Global trash goes to 318. */
+	    if (row_count_total > row_limit) {
 		D_(D_CRIT "row_count_total = %d @ ord %d, pat %d, row %d; ending scan", row_count_total, ord, pat, row);
 		goto end_module;
 	    }
 
-	    if (!loop_num && m->scan_cnt[ord][row]) {
+	    if (!loop_num && !line_jump && m->scan_cnt[ord][row]) {
 		row_count--;
 		goto end_module;
 	    }
@@ -237,12 +263,14 @@ static int scan_module(struct context_data *ctx, int ep, int chain)
 	    }
 
 	    pdelay = 0;
+	    line_jump = 0;
 
 	    for (chn = 0; chn < mod->chn; chn++) {
-		if (row >= mod->xxt[mod->xxp[pat]->index[chn]]->rows)
+		if (row >= tracks[chn]->rows)
 		    continue;
 
-		event = &EVENT(mod->xxo[ord], chn, row);
+		//event = &EVENT(mod->xxo[ord], chn, row);
+		event = &tracks[chn]->event[row];
 
 		f1 = event->fxt;
 		p1 = event->fxp;
@@ -290,23 +318,23 @@ static int scan_module(struct context_data *ctx, int ep, int chain)
 		    }
 		}
 
-		if ((f1 == FX_SPEED && p1) || (f2 == FX_SPEED && p2)) {
-		    parm = (f1 == FX_SPEED) ? p1 : p2;
+		/* Some formats can have two FX_SPEED effects, and both need
+		 * to be checked. Slot 2 is currently handled first. */
+		for (i = 0; i < 2; i++) {
+		    parm = i ? p1 : p2;
+		    if ((i ? f1 : f2) != FX_SPEED || parm == 0)
+			continue;
 		    frame_count += row_count * speed;
 		    row_count = 0;
-		    if (parm) {
-			if (HAS_QUIRK(QUIRK_NOBPM) || p->flags & XMP_FLAGS_VBLANK || parm < 0x20) {
-			    if (parm > 0) {
-			        speed = parm;
+		    if (HAS_QUIRK(QUIRK_NOBPM) || p->flags & XMP_FLAGS_VBLANK || parm < 0x20) {
+			speed = parm;
 #ifndef LIBXMP_CORE_PLAYER
-			        st26_speed = 0;
+			st26_speed = 0;
 #endif
-                            }
-			} else {
-			    time += m->time_factor * frame_count * base_time / bpm;
-			    frame_count = 0;
-			    bpm = parm;
-			}
+		    } else {
+			time += m->time_factor * frame_count * base_time / bpm;
+			frame_count = 0;
+			bpm = parm;
 		    }
 		}
 
@@ -327,6 +355,39 @@ static int scan_module(struct context_data *ctx, int ep, int chain)
 			st26_speed = MSN(p1);
 		    }
 		}
+
+		/* FAR tempo processing */
+
+		if (f1 == FX_FAR_TEMPO || f1 == FX_FAR_F_TEMPO) {
+			int far_speed, far_bpm, fine_change = 0;
+			if (f1 == FX_FAR_TEMPO) {
+				if (MSN(p1)) {
+					far_tempo_mode = MSN(p1) - 1;
+				} else {
+					far_tempo_coarse = LSN(p1);
+				}
+			}
+			if (f1 == FX_FAR_F_TEMPO) {
+				if (MSN(p1)) {
+					far_tempo_fine += MSN(p1);
+					fine_change = MSN(p1);
+				} else if (LSN(p1)) {
+					far_tempo_fine -= LSN(p1);
+					fine_change = -LSN(p1);
+				} else {
+					far_tempo_fine = 0;
+				}
+			}
+			if (libxmp_far_translate_tempo(far_tempo_mode, fine_change,
+			    far_tempo_coarse, &far_tempo_fine, &far_speed, &far_bpm) == 0) {
+				frame_count += row_count * speed;
+				row_count = 0;
+				time += m->time_factor * frame_count * base_time / bpm;
+				frame_count = 0;
+				speed = far_speed;
+				bpm = far_bpm;
+			}
+		}
 #endif
 
 		if ((f1 == FX_S3M_SPEED && p1) || (f2 == FX_S3M_SPEED && p2)) {
@@ -343,7 +404,7 @@ static int scan_module(struct context_data *ctx, int ep, int chain)
 
 		if ((f1 == FX_S3M_BPM && p1) || (f2 == FX_S3M_BPM && p2)) {
 		    parm = (f1 == FX_S3M_BPM) ? p1 : p2;
-		    if (parm >= 0x20) {
+		    if (parm >= XMP_MIN_BPM) {
 			frame_count += row_count * speed;
 			row_count = 0;
 			time += m->time_factor * frame_count * base_time / bpm;
@@ -417,6 +478,19 @@ static int scan_module(struct context_data *ctx, int ep, int chain)
 		    last_row = 0;
 		}
 
+#ifndef LIBXMP_CORE_PLAYER
+		/* Archimedes line jump */
+		if (f1 == FX_LINE_JUMP || f2 == FX_LINE_JUMP) {
+		    /* Don't set order if preceded by jump or break. */
+		    if (last_row > 0)
+			ord2 = ord;
+		    parm = (f1 == FX_LINE_JUMP) ? p1 : p2;
+		    break_row = parm;
+		    last_row = 0;
+		    line_jump = 1;
+		}
+#endif
+
 		if (f1 == FX_EXTENDED || f2 == FX_EXTENDED) {
 		    parm = (f1 == FX_EXTENDED) ? p1 : p2;
 
@@ -514,6 +588,56 @@ end_module:
     return (time + m->time_factor * frame_count * base_time / bpm);
 }
 
+static void reset_scan_data(struct context_data *ctx)
+{
+	int i;
+	for (i = 0; i < XMP_MAX_MOD_LENGTH; i++) {
+		ctx->m.xxo_info[i].time = -1;
+	}
+	memset(ctx->p.sequence_control, 0xff, XMP_MAX_MOD_LENGTH);
+}
+
+#ifndef LIBXMP_CORE_PLAYER
+static void compare_vblank_scan(struct context_data *ctx)
+{
+	/* Calculate both CIA and VBlank time for certain long MODs
+	 * and pick the more likely (i.e. shorter) one. The same logic
+	 * works regardless of the initial mode selected--either way,
+	 * the wrong timing mode usually makes modules MUCH longer. */
+	struct player_data *p = &ctx->p;
+	struct module_data *m = &ctx->m;
+	struct ord_data *info_backup;
+	struct scan_data scan_backup;
+	unsigned char ctrl_backup[256];
+
+	if ((info_backup = (struct ord_data *)malloc(sizeof(m->xxo_info))) != NULL) {
+		/* Back up the current info to avoid a third scan. */
+		scan_backup = p->scan[0];
+		memcpy(info_backup, m->xxo_info, sizeof(m->xxo_info));
+		memcpy(ctrl_backup, p->sequence_control,
+			sizeof(p->sequence_control));
+
+		reset_scan_data(ctx);
+
+		m->quirk ^= QUIRK_NOBPM;
+		p->scan[0].time = scan_module(ctx, 0, 0);
+
+		D_(D_INFO "%-6s %dms", !HAS_QUIRK(QUIRK_NOBPM)?"VBlank":"CIA", scan_backup.time);
+		D_(D_INFO "%-6s %dms",  HAS_QUIRK(QUIRK_NOBPM)?"VBlank":"CIA", p->scan[0].time);
+
+		if (p->scan[0].time >= scan_backup.time) {
+			m->quirk ^= QUIRK_NOBPM;
+			p->scan[0] = scan_backup;
+			memcpy(m->xxo_info, info_backup, sizeof(m->xxo_info));
+			memcpy(p->sequence_control, ctrl_backup,
+				sizeof(p->sequence_control));
+		}
+
+		free(info_backup);
+	}
+}
+#endif
+
 int libxmp_get_sequence(struct context_data *ctx, int ord)
 {
 	struct player_data *p = &ctx->p;
@@ -525,23 +649,35 @@ int libxmp_scan_sequences(struct context_data *ctx)
 	struct player_data *p = &ctx->p;
 	struct module_data *m = &ctx->m;
 	struct xmp_module *mod = &m->mod;
+	struct scan_data *s;
 	int i, ep;
 	int seq;
 	unsigned char temp_ep[XMP_MAX_MOD_LENGTH];
 
+	s = (struct scan_data *) realloc(p->scan, MAX(1, mod->len) * sizeof(struct scan_data));
+	if (!s) {
+		D_(D_CRIT "failed to allocate scan data");
+		return -1;
+	}
+	p->scan = s;
+
 	/* Initialize order data to prevent overwrite when a position is used
 	 * multiple times at different starting points (see janosik.xm).
 	 */
-	for (i = 0; i < XMP_MAX_MOD_LENGTH; i++) {
-		m->xxo_info[i].time = -1;
-	}
+	reset_scan_data(ctx);
 
 	ep = 0;
-	memset(p->sequence_control, 0xff, XMP_MAX_MOD_LENGTH);
 	temp_ep[0] = 0;
 	p->scan[0].time = scan_module(ctx, ep, 0);
 	seq = 1;
 
+#ifndef LIBXMP_CORE_PLAYER
+	if (m->compare_vblank && !(p->flags & XMP_FLAGS_VBLANK) &&
+	    p->scan[0].time >= VBLANK_TIME_THRESHOLD) {
+		compare_vblank_scan(ctx);
+	}
+#endif
+
 	if (p->scan[0].time < 0) {
 		D_(D_CRIT "scan was not able to find any valid orders");
 		return -1;
@@ -567,6 +703,12 @@ int libxmp_scan_sequences(struct context_data *ctx)
 		}
 	}
 
+	if (seq < mod->len) {
+		s = (struct scan_data *) realloc(p->scan, seq * sizeof(struct scan_data));
+		if (s != NULL) {
+			p->scan = s;
+		}
+	}
 	m->num_sequences = seq;
 
 	/* Now place entry points in the public accessible array */
@@ -575,6 +717,5 @@ int libxmp_scan_sequences(struct context_data *ctx)
 		m->seq_data[i].duration = p->scan[i].time;
 	}
 
-
 	return 0;
 }
